from os.path import basename, dirname
from sys import argv, exit
from os import getcwd

from PyQt4.QtCore import SIGNAL
from PyQt4.QtCore import Qt
from PyQt4.QtGui import (QApplication, QDialog,
                         QTableWidgetItem, QFileDialog, QMessageBox)

from hachoir.metadata.qt.dialog_ui import Ui_Form
from hachoir.core import config
from hachoir.core.error import HACHOIR_ERRORS
from hachoir.core.cmd_line import unicodeFilename
from hachoir.parser import createParser
from hachoir.metadata import extractMetadata
from hachoir.metadata.metadata import MultipleMetadata


config.quiet = True


class CustomTableWidgetItem(QTableWidgetItem):

    def __init__(self, parent=None):
        super(CustomTableWidgetItem, self).__init__(parent)
        self.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)


class MetadataError(Exception):

    def __init__(self, message):
        self.unicode_message = message
        bytes_message = message.encode("ASCII", "replace")
        Exception.__init__(self, bytes_message)

    def __unicode__(self):
        return self.unicode_message


class File:

    def __init__(self, filename, realname):
        self.name = filename
        if realname:
            self.realname = realname
        else:
            self.realname = filename
        try:
            self.parser = createParser(self.name, self.realname)
        except HACHOIR_ERRORS as err:
            raise MetadataError("Parser error: %s" % str(err))
        if not self.parser:
            raise MetadataError("Unable to parse the file: %s" % self.name)
        try:
            self.metadata = extractMetadata(self.parser)
        except HACHOIR_ERRORS as err:
            raise MetadataError("Metadata extraction error: %s" % str(err))
        if not self.metadata:
            file_type = self.parser.mime_type
            raise MetadataError(
                "Unable to extract metadata from file of type %s" % file_type)


class Metadata(QDialog, Ui_Form):

    def __init__(self, application):
        QDialog.__init__(self)
        self.application = application
        self.files = {}
        self.setupWindow()
        self.current_dir = dirname(getcwd())

    def setupWindow(self):
        self.setupUi(self)
        self.connect(self.open_button, SIGNAL("clicked()"), self.open)
        self.connect(self.quit_button, SIGNAL("clicked()"), self.quit)
        self.connect(
            self.files_combo,
            SIGNAL("currentIndexChanged(const QString&)"),
            self.changeFile)
        self.metadata_table.horizontalHeader().hide()
        self.metadata_table.verticalHeader().hide()

    def open(self):
        filename = QFileDialog.getOpenFileName(
            self,  # parent
            "Choose a file to open",  # caption
            self.current_dir,  # dir name
            "",  # filter
            "Any type"  # filter set
        )
        if not filename:
            return
        filename = str(filename)
        self.current_dir = dirname(filename)
        self.addFile(filename, change=True)

    def quit(self):
        self.application.quit()

    def fillList(self, file):
        table = self.metadata_table
        metadata = file.metadata
        groups = [metadata]
        if isinstance(metadata, MultipleMetadata):
            groups.extend(list(metadata.iterGroups()))
        total = 0
        for index, metadata in enumerate(groups):
            group_name = metadata.header
            metadata = [data for data in metadata if data.values]
            metadata.sort()
            if 0 < index:
                metadata.insert(0, group_name)
            groups[index] = metadata
            total += len(metadata)
        table.clear()
        table.setColumnCount(2)
        table.setRowCount(total)
        row = 0
        for metadata in groups:
            for data in metadata:
                if isinstance(data, str):
                    table.setItem(row, 0, CustomTableWidgetItem("-- group --"))
                    table.setItem(row, 1, CustomTableWidgetItem(data))
                    row += 1
                    continue
                title = data.description
                for item in data.values:
                    value = item.text
                    table.setItem(row, 0, CustomTableWidgetItem(title))
                    table.setItem(row, 1, CustomTableWidgetItem(value))
                    row += 1
        table.resizeColumnsToContents()

    def changeFile(self, name):
        name = str(name)
        file = self.files[name]
        self.fillList(file)

    def error(self, message):
        QMessageBox.warning(self, "Metadata error", message)

    def addFile(self, filename, realname=None, change=False):
        try:
            file = File(filename, realname)
        except MetadataError as err:
            errmsg = str(err)
            self.error(errmsg)
            return
        name = basename(file.name)
        self.files[name] = file
        self.files_combo.addItem(name)
        if change:
            index = self.files_combo.count() - 1
            self.files_combo.setCurrentIndex(index)


def main():
    app = QApplication(argv)
    metadata = Metadata(app)
    for filename in argv[1:]:
        realname = filename
        filename = unicodeFilename(filename)
        metadata.addFile(filename, realname)
    metadata.show()
    exitcode = app.exec_()
    exit(exitcode)



from hachoir.field import available_types
from hachoir.wx.field_view.mutator import convert_field
from hachoir.wx.field_view.stubs import can_convert


class core_type_menu_imp_t:
    def __init__(self):
        self.cur_field = None

    def on_core_type_menu_ready(self, dispatcher, view):
        assert view is not None
        self.view = view

    def on_type_selected(self, id):
        convert_field(self.cur_field, self.view.get_type_name(id))
        self.dispatcher.trigger('field_modified', self.cur_field)

    def on_field_selected(self, dispatcher, field):
        self.cur_field = field

        self.view.clear()
        for type in available_types:
            if can_convert(field, type) and field.__class__ is not type:
                self.view.add_type(type.__name__)



"""
iPod iTunesDB parser.

Documentation:
- http://ipl.derpapst.org/wiki/ITunesDB/iTunesDB_File

Author: Romain HERAULT
Creation date: 19 august 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt8, UInt16, UInt32, Int32, UInt64, TimestampMac32,
                           String, Float32, NullBytes, Enum, RawBytes)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.tools import humanDuration
from hachoir.core.text_handler import displayHandler, filesizeHandler

list_order = {
    1: "playlist order (manual sort order)",
    2: "???",
    3: "songtitle",
    4: "album",
    5: "artist",
    6: "bitrate",
    7: "genre",
    8: "kind",
    9: "date modified",
    10: "track number",
    11: "size",
    12: "time",
    13: "year",
    14: "sample rate",
    15: "comment",
    16: "date added",
    17: "equalizer",
    18: "composer",
    19: "???",
    20: "play count",
    21: "last played",
    22: "disc number",
    23: "my rating",
    24: "release date",
    25: "BPM",
    26: "grouping",
    27: "category",
    28: "description",
    29: "show",
    30: "season",
    31: "episode number"
}


class DataObject(FieldSet):
    type_name = {
        1: "Title",
        2: "Location",
        3: "Album",
        4: "Artist",
        5: "Genre",
        6: "Filetype",
        7: "EQ Setting",
        8: "Comment",
        9: "Category",
        12: "Composer",
        13: "Grouping",
        14: "Description text",
        15: "Podcast Enclosure URL",
        16: "Podcast RSS URL",
        17: "Chapter data",
        18: "Subtitle",
        19: "Show (for TV Shows only)",
        20: "Episode",
        21: "TV Network",
        22: "Album-Artist",
        23: "Artist for Sorting",
        24: "List of keywords pretaining track",
        25: "Locale for TV show(?)",
        27: "Title for Sorting",
        28: "Album for Sorting",
        29: "Album-Artist for Sorting",
        30: "Composer for Sorting",
        31: "Show for Sorting",
        # 32: "Unknown binary field for video tracks",
        50: "Smart Playlist Data",
        51: "Smart Playlist Rules",
        52: "Library Playlist Index",
        53: "Library Playlist Index letter in jump table",
        100: "Ccolumn Sizing Info as well as an order indicator in playlists.",
        102: "For iPhone",
        200: "Album name (for album descriptions)",
        201: "Album artist (for album descriptions)",
        202: "Album sort artist (for album descriptions)",
        203: "Podcast URL in Album List",
        204: "TV Show in Album List"
    }

    mhod52_sort_index_type_name = {
        3: "Title",
        4: "Album, then Disk/Tracknumber, then Title",
        5: "Artist, then Album, then Disc/Tracknumber, then Title",
        7: "Genre, then Artist, then Album, then Disc/Tracknumber, then Title",
        8: "Composer, then Title"
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value * 8

    def createFields(self):
        yield String(self, "header_id", 4, "Data Object Header Markup (\"mhod\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield Enum(UInt32(self, "type", "type"), self.type_name)
        if (self["type"].value == 15) or (self["type"].value == 16):
            yield UInt32(self, "unknown[]")
            yield UInt32(self, "unknown[]")
            yield String(self, "string", self._size // 8 - self["header_length"].value, "String Data", charset="UTF-8")
        elif (self["type"].value == 52):
            yield UInt32(self, "unknown[]", "unk1")
            yield UInt32(self, "unknown[]", "unk2")
            yield Enum(UInt32(self, "sort_index_type", "Sort Index Type"), self.mhod52_sort_index_type_name)
            yield UInt32(self, "entry_count", "Entry Count")
            indexes_size = self["entry_count"].value * 4
            padding_offset = self["entry_length"].value - indexes_size
            padding = self.seekByte(padding_offset, "header padding")
            if padding:
                yield padding
            for i in range(self["entry_count"].value):
                yield UInt32(self, "index[" + str(i) + "]", "Index of the " + str(i) + "nth mhit")
        elif (self["type"].value < 15) or (self["type"].value > 17) or (self["type"].value >= 200):
            yield UInt32(self, "unknown[]")
            yield UInt32(self, "unknown[]")
            yield UInt32(self, "position", "Position")
            yield UInt32(self, "length", "String Length in bytes")
            yield UInt32(self, "unknown[]")
            yield UInt32(self, "unknown[]")
            yield String(self, "string", self["length"].value, "String Data", charset="UTF-16-LE")
        else:
            padding = self.seekByte(
                self["header_length"].value, "header padding")
            if padding:
                yield padding
        padding = self.seekBit(self._size, "entry padding")
        if padding:
            yield padding


class TrackItem(FieldSet):
    x1_type_name = {
        0: "AAC or CBR MP3",
        1: "VBR MP3"
    }
    x2_type_name = {
        0: "AAC",
        1: "MP3"
    }
    media_type_name = {
        0x00: "Audio/Video",
        0x01: "Audio",
        0x02: "Video",
        0x04: "Podcast",
        0x06: "Video Podcast",
        0x08: "Audiobook",
        0x20: "Music Video",
        0x40: "TV Show",
        0X60: "TV Show (Music lists)",
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value * 8

    def createFields(self):
        yield String(self, "header_id", 4, "Track Item Header Markup (\"mhit\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield UInt32(self, "string_number", "Number of Strings")
        yield UInt32(self, "unique_id", "Unique ID")
        yield UInt32(self, "visible_tag", "Visible Tag")
        yield String(self, "file_type", 4, "File Type")
        yield Enum(UInt8(self, "x1_type", "Extended Type 1"), self.x1_type_name)
        yield Enum(UInt8(self, "x2_type", "Extended type 2"), self.x2_type_name)
        yield UInt8(self, "compilation_flag", "Compilation Flag")
        yield UInt8(self, "rating", "Rating")
        yield TimestampMac32(self, "last_modified", "Time of the last modification of the track")
        yield filesizeHandler(UInt32(self, "size", "Track size in bytes"))
        yield displayHandler(UInt32(self, "length", "Track length in milliseconds"), humanDuration)
        yield UInt32(self, "track_number", "Number of this track")
        yield UInt32(self, "total_track", "Total number of tracks")
        yield UInt32(self, "year", "Year of the track")
        yield UInt32(self, "bitrate", "Bitrate")
        yield UInt32(self, "samplerate", "Sample Rate")
        yield UInt32(self, "volume", "volume")
        yield UInt32(self, "start_time", "Start playing at, in milliseconds")
        yield UInt32(self, "stop_time", "Stop playing at,  in milliseconds")
        yield UInt32(self, "soundcheck", "SoundCheck preamp")
        yield UInt32(self, "playcount_1", "Play count of the track")
        yield UInt32(self, "playcount_2", "Play count of the track when last synced")
        yield TimestampMac32(self, "last_played_time", "Time the song was last played")
        yield UInt32(self, "disc_number", "disc number in multi disc sets")
        yield UInt32(self, "total_discs", "Total number of discs in the disc set")
        yield UInt32(self, "userid", "User ID in the DRM scheme")
        yield TimestampMac32(self, "added_date", "Date when the item was added")
        yield UInt32(self, "bookmark_time", "Bookmark time for AudioBook")
        yield UInt64(self, "dbid", "Unique DataBase ID for the song (identical in mhit and in mhii)")
        yield UInt8(self, "checked", "song is checked")
        yield UInt8(self, "application_rating", "Last Rating before change")
        yield UInt16(self, "BPM", "BPM of the track")
        yield UInt16(self, "artwork_count", "number of artworks for this item")
        yield UInt16(self, "unknown[]")
        yield UInt32(self, "artwork_size", "Total size of artworks in bytes")
        yield UInt32(self, "unknown[]")
        yield Float32(self, "sample_rate_2", "Sample Rate express in float")
        yield UInt32(self, "released_date", "Date of release in Music Store or in Podcast")
        yield UInt16(self, "unknown[]")
        yield UInt16(self, "explicit_flag[]", "Explicit flag")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "skip_count[]", "Skip Count")
        yield TimestampMac32(self, "last_skipped", "Date when the item was last skipped")
        yield UInt8(self, "has_artwork", "0x01 for track with artwork, 0x02 otherwise")
        yield UInt8(self, "skip_wen_shuffling", "Skip that track when shuffling")
        yield UInt8(self, "remember_playback_position", "Remember playback position")
        yield UInt8(self, "flag4", "Flag 4")
        yield UInt64(self, "dbid2", "Unique DataBase ID for the song (identical as above)")
        yield UInt8(self, "lyrics_flag", "Lyrics Flag")
        yield UInt8(self, "movie_file_flag", "Movie File Flag")
        yield UInt8(self, "played_mark", "Track has been played")
        yield UInt8(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "pregap[]", "Number of samples of silence before the song starts")
        yield UInt64(self, "sample_count", "Number of samples in the song (only for WAV and AAC files)")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "postgap[]", "Number of samples of silence at the end of the song")
        yield UInt32(self, "unknown[]")
        yield Enum(UInt32(self, "media_type", "Media Type for video iPod"), self.media_type_name)
        yield UInt32(self, "season_number", "Season Number")
        yield UInt32(self, "episode_number", "Episode Number")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "gapless_data[]", "The size in bytes from first Sync Frame until the 8th before the last frame.")
        yield UInt32(self, "unknown[]")
        yield UInt16(self, "gaplessTrackFlag[]", "1 if track has gapless data")
        yield UInt16(self, "gaplessAlbumFlag[]", "1 if track uses crossfading in iTunes")
        yield RawBytes(self, "unknown[]", 20)
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt16(self, "unknown[]")
        yield UInt16(self, "album_id[]", "Album ID (used to link tracks with MHIAs)")
        yield RawBytes(self, "unknown[]", 52)
        yield UInt32(self, "mhii_link[]", "Artwork ID (used to link tracks with MHIIs)")
        padding = self.seekByte(self["header_length"].value, "header padding")
        if padding:
            yield padding

        # while ((self.stream.readBytes(0, 4) == b'mhod') and
        # ((self.current_size//8) < self["entry_length"].value)):
        for i in range(self["string_number"].value):
            yield DataObject(self, "data[]")
        padding = self.seekBit(self._size, "entry padding")
        if padding:
            yield padding


class TrackList(FieldSet):

    def createFields(self):
        yield String(self, "header_id", 4, "Track List Header Markup (\"mhlt\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "track_number", "Number of Tracks")

        padding = self.seekByte(self["header_length"].value, "header padding")
        if padding:
            yield padding

        for i in range(self["track_number"].value):
            yield TrackItem(self, "track[]")


class PlaylistItem(FieldSet):

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value * 8

    def createFields(self):
        yield String(self, "header_id", 4, "Playlist Item Header Markup (\"mhip\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield UInt32(self, "data_object_child_count", "Number of Child Data Objects")
        yield UInt32(self, "podcast_grouping_flag", "Podcast Grouping Flag")
        yield UInt32(self, "group_id", "Group ID")
        yield UInt32(self, "track_id", "Track ID")
        yield TimestampMac32(self, "timestamp", "Song Timestamp")
        yield UInt32(self, "podcast_grouping_ref", "Podcast Grouping Reference")
        padding = self.seekByte(self["header_length"].value, "header padding")
        if padding:
            yield padding

        for i in range(self["data_object_child_count"].value):
            yield DataObject(self, "mhod[]")


class Playlist(FieldSet):
    is_master_pl_name = {
        0: "Regular playlist",
        1: "Master playlist"
    }

    is_podcast_name = {
        0: "Normal Playlist List",
        1: "Podcast Playlist List"
    }

    list_sort_order_name = {
        1: "Manual Sort Order",
        2: "???",
        3: "Song Title",
        4: "Album",
        5: "Artist",
        6: "Bitrate",
        7: "Genre",
        8: "Kind",
        9: "Date Modified",
        10: "Track Number",
        11: "Size",
        12: "Time",
        13: "Year",
        14: "Sample Rate",
        15: "Comment",
        16: "Date Added",
        17: "Equalizer",
        18: "Composer",
        19: "???",
        20: "Play Count",
        21: "Last Played",
        22: "Disc Number",
        23: "My Rating",
        24: "Release Date",
        25: "BPM",
        26: "Grouping",
        27: "Category",
        28: "Description",
        29: "Show",
        30: "Season",
        31: "Episode Number"
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value * 8

    def createFields(self):
        yield String(self, "header_id", 4, "Playlist Header Markup (\"mhyp\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield UInt32(self, "data_object_child_count", "Number of Child Data Objects")
        yield UInt32(self, "playlist_count", "Number of Playlist Items")
        yield Enum(UInt8(self, "type", "Normal or master playlist?"), self.is_master_pl_name)
        yield UInt8(self, "XXX1", "XXX1")
        yield UInt8(self, "XXX2", "XXX2")
        yield UInt8(self, "XXX3", "XXX3")
        yield TimestampMac32(self, "creation_date", "Date when the playlist was created")
        yield UInt64(self, "playlistid", "Persistent Playlist ID")
        yield UInt32(self, "unk3", "unk3")
        yield UInt16(self, "string_mhod_count", "Number of string MHODs for this playlist")
        yield Enum(UInt16(self, "is_podcast", "Playlist or Podcast List?"), self.is_podcast_name)
        yield Enum(UInt32(self, "sort_order", "Playlist Sort Order"), self.list_sort_order_name)

        padding = self.seekByte(self["header_length"].value, "entry padding")
        if padding:
            yield padding

        for i in range(self["data_object_child_count"].value):
            yield DataObject(self, "mhod[]")

        for i in range(self["playlist_count"].value):
            yield PlaylistItem(self, "playlist_item[]")


class PlaylistList(FieldSet):

    def createFields(self):
        yield String(self, "header_id", 4, "Playlist List Header Markup (\"mhlp\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "playlist_number", "Number of Playlists")

        padding = self.seekByte(self["header_length"].value, "header padding")
        if padding:
            yield padding

        for i in range(self["playlist_number"].value):
            yield Playlist(self, "playlist[]")


class Album(FieldSet):

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value * 8

    def createFields(self):
        yield String(self, "header_id", 4, "Album Item Header Markup (\"mhia\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield UInt32(self, "data_object_child_count", "Number of Child Data Objects")
        yield UInt16(self, "unknow[]")
        yield UInt16(self, "album_id[]", "Album ID")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")

        padding = self.seekByte(self["header_length"].value, "entry padding")
        if padding:
            yield padding

        for i in range(self["data_object_child_count"].value):
            yield DataObject(self, "mhod[]")


class AlbumList(FieldSet):

    def createFields(self):
        yield String(self, "header_id", 4, "Album List Header Markup (\"mhla\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "album_number", "Number of Albums")

        padding = self.seekByte(self["header_length"].value, "header padding")
        if padding:
            yield padding

        for i in range(self["album_number"].value):
            yield Album(self, "album[]")


class DataSet(FieldSet):
    type_name = {
        1: "Track List",
        2: "Play List",
        3: "Podcast List",
        4: "Album List"
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value * 8

    def createFields(self):
        yield String(self, "header_id", 4, "DataSet Header Markup (\"mhsd\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield Enum(UInt32(self, "type", "type"), self.type_name)
        padding = self.seekByte(self["header_length"].value, "header_raw")
        if padding:
            yield padding
        if self["type"].value == 1:
            yield TrackList(self, "tracklist[]")
        if self["type"].value == 2:
            yield PlaylistList(self, "playlist_list[]")
        if self["type"].value == 3:
            yield PlaylistList(self, "podcast_list[]")
        if self["type"].value == 4:
            yield AlbumList(self, "album_list[]")
        padding = self.seekBit(self._size, "entry padding")
        if padding:
            yield padding


class DataBase(FieldSet):

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value * 8

#    def createFields(self):


class ITunesDBFile(Parser):
    PARSER_TAGS = {
        "id": "itunesdb",
        "category": "audio",
        "min_size": 44 * 8,
        "magic": ((b'mhbd', 0),),
        "description": "iPod iTunesDB file"
    }

    endian = LITTLE_ENDIAN

    def validate(self):
        return self.stream.readBytes(0, 4) == b'mhbd'

    def createFields(self):
        yield String(self, "header_id", 4, "DataBase Header Markup (\"mhbd\")", charset="ISO-8859-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "version_number", "Version Number")
        yield UInt32(self, "child_number", "Number of Children")
        yield UInt64(self, "id", "ID for this database")
        yield UInt16(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield UInt64(self, "unknown[]")
        yield UInt16(self, "unknown[]")
        yield UInt16(self, "hashing_scheme[]", "Algorithm used to calculate the database hash")
        yield NullBytes(self, "unknown[]", 20)
        yield String(self, "language_id", 2, "Language ID")
        yield UInt64(self, "persistent_id", "Library Persistent ID")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        yield RawBytes(self, "hash[]", 20)
        yield Int32(self, "timezone_offset[]", "Timezone offset in seconds")
        yield UInt16(self, "unknown[]")
        yield RawBytes(self, "iphone_hash[]", 45)
        size = self["header_length"].value - self.current_size / 8
        if size > 0:
            yield NullBytes(self, "padding", size)
        for i in range(self["child_number"].value):
            yield DataSet(self, "dataset[]")
        padding = self.seekByte(self["entry_length"].value, "entry padding")
        if padding:
            yield padding

    def createContentSize(self):
        return self["entry_length"].value * 8



"""
Parser of FastTrackerII Extended Module (XM) version 1.4

Documents:
- Modplug source code (file modplug/soundlib/Load_mod.cpp)
  http://sourceforge.net/projects/modplug
- Dumb source code (files include/dumb.h and src/it/readmod.c
  http://dumb.sf.net/
- Documents on "MOD" format on Wotsit
  http://www.wotsit.org

Compressed formats (i.e. starting with "PP20" or having "PACK" as type
are not handled. Also NoiseTracker's NST modules aren't handled, although
it might be possible: no file format and 15 samples

Author: Christophe GISQUET <christophe.gisquet@free.fr>
Creation: 18th February 2007
"""

from math import log10
from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           Bits, UInt16, UInt8,
                           RawBytes, String, GenericVector)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import textHandler

# Old NoiseTracker 15-samples modules can have anything here.
MODULE_TYPE = {
    b"M.K.": ("Noise/Pro-Tracker", 4),
    b"M!K!": ("Noise/Pro-Tracker", 4),
    b"M&K&": ("Noise/Pro-Tracker", 4),
    b"RASP": ("StarTrekker", 4),
    b"FLT4": ("StarTrekker", 4),
    b"FLT8": ("StarTrekker", 8),
    b"6CHN": ("FastTracker", 6),
    b"8CHN": ("FastTracker", 8),
    b"CD81": ("Octalyser", 8),
    b"OCTA": ("Octalyser", 8),
    b"FA04": ("Digital Tracker", 4),
    b"FA06": ("Digital Tracker", 6),
    b"FA08": ("Digital Tracker", 8),
}


def getFineTune(val):
    return ("0", "1", "2", "3", "4", "5", "6", "7", "8",
            "-8", "-7", "-6", "-5", "-4", "-3", "-2", "-1")[val.value]


def getVolume(val):
    return "%.1f dB" % (20.0 * log10(val.value / 64.0))


class SampleInfo(FieldSet):
    static_size = 30 * 8

    def createFields(self):
        yield String(self, "name", 22, strip='\0')
        yield UInt16(self, "sample_count")
        yield textHandler(UInt8(self, "fine_tune"), getFineTune)
        yield textHandler(UInt8(self, "volume"), getVolume)
        yield UInt16(self, "loop_start", "Loop start offset in samples")
        yield UInt16(self, "loop_len", "Loop length in samples")

    def createValue(self):
        return self["name"].value


class Header(FieldSet):
    static_size = 1084 * 8

    def createFields(self):
        yield String(self, "name", 20, strip='\0')
        yield GenericVector(self, "samples", 31, SampleInfo, "info")
        yield UInt8(self, "length")
        yield UInt8(self, "played_patterns_count")
        yield GenericVector(self, "patterns", 128, UInt8, "position")
        yield String(self, "type", 4)

    def getNumChannels(self):
        return MODULE_TYPE[self["type"].value][1]


class Note(FieldSet):
    static_size = 8 * 4

    def createFields(self):
        yield Bits(self, 4, "note_hi_nibble")
        yield Bits(self, 12, "period")
        yield Bits(self, 4, "note_low_nibble")
        yield Bits(self, 4, "effect")
        yield UInt8(self, "parameter")


class Row(FieldSet):

    def __init__(self, parent, name, channels, desc=None):
        FieldSet.__init__(self, parent, name, description=desc)
        self.channels = channels
        self._size = 8 * self.channels * 4

    def createFields(self):
        for index in range(self.channels):
            yield Note(self, "note[]")


class Pattern(FieldSet):

    def __init__(self, parent, name, channels, desc=None):
        FieldSet.__init__(self, parent, name, description=desc)
        self.channels = channels
        self._size = 64 * 8 * self.channels * 4

    def createFields(self):
        for index in range(64):
            yield Row(self, "row[]", self.channels)


class AmigaModule(Parser):
    PARSER_TAGS = {
        "id": "mod",
        "category": "audio",
        "file_ext": ("mod", "nst", "wow", "oct", "sd0"),
        "mime": ('audio/mod', 'audio/x-mod', 'audio/mod', 'audio/x-mod'),
        "min_size": 1084 * 8,
        "description": "Uncompressed amiga module"
    }
    endian = BIG_ENDIAN

    def validate(self):
        modtype = self.stream.readBytes(1080 * 8, 4)
        if modtype not in MODULE_TYPE:
            return "Invalid module type %a" % modtype
        self.createValue = lambda modtype: "%s module, %u channels" % MODULE_TYPE[
            modtype]
        return True

    def createFields(self):
        header = Header(self, "header")
        yield header
        channels = header.getNumChannels()

        # Number of patterns
        patterns = 0
        for index in range(128):
            patterns = max(patterns,
                           header["patterns/position[%u]" % index].value)
        patterns += 1

        # Yield patterns
        for index in range(patterns):
            yield Pattern(self, "pattern[]", channels)

        # Yield samples
        for index in range(31):
            count = header["samples/info[%u]/sample_count" % index].value
            if count:
                self.info("Yielding sample %u: %u samples" % (index, count))
                yield RawBytes(self, "sample_data[]", 2 * count,
                               "Sample %u" % index)



"""
Mapsforge map file parser (for version 3 files).

Author: Oliver Gerlich

References:
- http://code.google.com/p/mapsforge/wiki/SpecificationBinaryMapFile
- http://mapsforge.org/
"""

from hachoir.parser import Parser
from hachoir.field import (Bit, Bits, UInt8, UInt16, UInt32, Int32, UInt64, String,
                           PaddingBits,
                           Enum, Field, FieldSet, SeekableFieldSet, RootSeekableFieldSet)
from hachoir.core.endian import BIG_ENDIAN


# micro-degrees factor:
UDEG = float(1000 * 1000)


CoordinateEncoding = {
    0: "single delta encoding",
    1: "double delta encoding",
}


class UIntVbe(Field):

    def __init__(self, parent, name, description=None):
        Field.__init__(self, parent, name, description=description)

        value = 0
        size = 0
        while True:
            byteValue = self._parent.stream.readBytes(
                self.absolute_address + (size * 8), 1)[0]

            haveMoreData = (byteValue & 0x80)
            value = value | ((byteValue & 0x7f) << (size * 7))
            size += 1
            assert size < 100, "UIntVBE is too large"

            if not haveMoreData:
                break

        self._size = size * 8
        self.createValue = lambda: value


class IntVbe(Field):

    def __init__(self, parent, name, description=None):
        Field.__init__(self, parent, name, description=description)

        value = 0
        size = 0
        shift = 0
        while True:
            byteValue = self._parent.stream.readBytes(
                self.absolute_address + (size * 8), 1)[0]

            haveMoreData = (byteValue & 0x80)
            if size == 0:
                isNegative = (byteValue & 0x40)
                value = (byteValue & 0x3f)
                shift += 6
            else:
                value = value | ((byteValue & 0x7f) << shift)
                shift += 7
            size += 1
            assert size < 100, "IntVBE is too large"

            if not haveMoreData:
                break

        if isNegative:
            value *= -1

        self._size = size * 8
        self.createValue = lambda: value


class VbeString(FieldSet):

    def createFields(self):
        yield UIntVbe(self, "length")
        yield String(self, "chars", self["length"].value, charset="UTF-8")

    def createDescription(self):
        return '(%d B) "%s"' % (self["length"].value, self["chars"].value)


class TagStringList(FieldSet):

    def createFields(self):
        yield UInt16(self, "num_tags")
        for i in range(self["num_tags"].value):
            yield VbeString(self, "tag[]")

    def createDescription(self):
        return "%d tag strings" % self["num_tags"].value


class ZoomIntervalCfg(FieldSet):

    def createFields(self):
        yield UInt8(self, "base_zoom_level")
        yield UInt8(self, "min_zoom_level")
        yield UInt8(self, "max_zoom_level")
        yield UInt64(self, "subfile_start")
        yield UInt64(self, "subfile_size")

    def createDescription(self):
        return "zoom level around %d (%d - %d)" % (self["base_zoom_level"].value,
                                                   self["min_zoom_level"].value, self["max_zoom_level"].value)


class TileIndexEntry(FieldSet):

    def createFields(self):
        yield Bit(self, "is_water_tile")
        yield Bits(self, "offset", 39)


class TileZoomTable(FieldSet):

    def createFields(self):
        yield UIntVbe(self, "num_pois")
        yield UIntVbe(self, "num_ways")

    def createDescription(self):
        return "%d POIs, %d ways" % (self["num_pois"].value, self["num_ways"].value)


class TileHeader(FieldSet):

    def __init__(self, parent, name, zoomIntervalCfg, **kw):
        FieldSet.__init__(self, parent, name, **kw)
        self.zoomIntervalCfg = zoomIntervalCfg

    def createFields(self):
        numLevels = int(self.zoomIntervalCfg[
                        "max_zoom_level"].value - self.zoomIntervalCfg["min_zoom_level"].value) + 1
        assert (numLevels < 50)
        for i in range(numLevels):
            yield TileZoomTable(self, "zoom_table_entry[]")
        yield UIntVbe(self, "first_way_offset")


class POIData(FieldSet):

    def createFields(self):
        if self["/have_debug"].value:
            yield String(self, "signature", 32)
            if not self['signature'].value.startswith("***POIStart"):
                raise ValueError

        yield IntVbe(self, "lat_diff")
        yield IntVbe(self, "lon_diff")
        yield Bits(self, "layer", 4)
        yield Bits(self, "num_tags", 4)

        for i in range(self["num_tags"].value):
            yield UIntVbe(self, "tag_id[]")

        yield Bit(self, "have_name")
        yield Bit(self, "have_house_number")
        yield Bit(self, "have_ele")
        yield PaddingBits(self, "pad[]", 5)

        if self["have_name"].value:
            yield VbeString(self, "name")
        if self["have_house_number"].value:
            yield VbeString(self, "house_number")
        if self["have_ele"].value:
            yield IntVbe(self, "ele")

    def createDescription(self):
        s = "POI"
        if self["have_name"].value:
            s += ' "%s"' % self["name"]["chars"].value
        s += " @ %f/%f" % (self["lat_diff"].value / UDEG,
                           self["lon_diff"].value / UDEG)
        return s


class SubTileBitmap(FieldSet):
    static_size = 2 * 8

    def createFields(self):
        for y in range(4):
            for x in range(4):
                yield Bit(self, "is_used[%d,%d]" % (x, y))


class WayProperties(FieldSet):

    def createFields(self):
        if self["/have_debug"].value:
            yield String(self, "signature", 32)
            if not self['signature'].value.startswith("---WayStart"):
                raise ValueError

        yield UIntVbe(self, "way_data_size")

        # WayProperties is split into an outer and an inner field, to allow
        # specifying data size for inner part:
        yield WayPropertiesInner(self, "inner", size=self["way_data_size"].value * 8)


class WayPropertiesInner(FieldSet):

    def createFields(self):
        yield SubTileBitmap(self, "sub_tile_bitmap")
        # yield Bits(self, "sub_tile_bitmap", 16)

        yield Bits(self, "layer", 4)
        yield Bits(self, "num_tags", 4)

        for i in range(self["num_tags"].value):
            yield UIntVbe(self, "tag_id[]")

        yield Bit(self, "have_name")
        yield Bit(self, "have_house_number")
        yield Bit(self, "have_ref")
        yield Bit(self, "have_label_position")
        yield Bit(self, "have_num_way_blocks")
        yield Enum(Bit(self, "coord_encoding"), CoordinateEncoding)
        yield PaddingBits(self, "pad[]", 2)

        if self["have_name"].value:
            yield VbeString(self, "name")
        if self["have_house_number"].value:
            yield VbeString(self, "house_number")
        if self["have_ref"].value:
            yield VbeString(self, "ref")
        if self["have_label_position"].value:
            yield IntVbe(self, "label_lat_diff")
            yield IntVbe(self, "label_lon_diff")
        numWayDataBlocks = 1
        if self["have_num_way_blocks"].value:
            yield UIntVbe(self, "num_way_blocks")
            numWayDataBlocks = self["num_way_blocks"].value

        for i in range(numWayDataBlocks):
            yield WayData(self, "way_data[]")

    def createDescription(self):
        s = "way"
        if self["have_name"].value:
            s += ' "%s"' % self["name"]["chars"].value
        return s


class WayData(FieldSet):

    def createFields(self):
        yield UIntVbe(self, "num_coord_blocks")
        for i in range(self["num_coord_blocks"].value):
            yield WayCoordBlock(self, "way_coord_block[]")


class WayCoordBlock(FieldSet):

    def createFields(self):
        yield UIntVbe(self, "num_way_nodes")
        yield IntVbe(self, "first_lat_diff")
        yield IntVbe(self, "first_lon_diff")

        for i in range(self["num_way_nodes"].value - 1):
            yield IntVbe(self, "lat_diff[]")
            yield IntVbe(self, "lon_diff[]")


class TileData(FieldSet):

    def __init__(self, parent, name, zoomIntervalCfg, **kw):
        FieldSet.__init__(self, parent, name, **kw)
        self.zoomIntervalCfg = zoomIntervalCfg

    def createFields(self):
        if self["/have_debug"].value:
            yield String(self, "signature", 32)
            if not self['signature'].value.startswith("###TileStart"):
                raise ValueError

        yield TileHeader(self, "tile_header", self.zoomIntervalCfg)

        numLevels = int(self.zoomIntervalCfg[
                        "max_zoom_level"].value - self.zoomIntervalCfg["min_zoom_level"].value) + 1
        for zoomLevel in range(numLevels):
            zoomTableEntry = self["tile_header"][
                "zoom_table_entry[%d]" % zoomLevel]
            for poiIndex in range(zoomTableEntry["num_pois"].value):
                yield POIData(self, "poi_data[%d,%d]" % (zoomLevel, poiIndex))

        for zoomLevel in range(numLevels):
            zoomTableEntry = self["tile_header"][
                "zoom_table_entry[%d]" % zoomLevel]
            for wayIndex in range(zoomTableEntry["num_ways"].value):
                yield WayProperties(self, "way_props[%d,%d]" % (zoomLevel, wayIndex))


class ZoomSubFile(SeekableFieldSet):

    def __init__(self, parent, name, zoomIntervalCfg, **kw):
        SeekableFieldSet.__init__(self, parent, name, **kw)
        self.zoomIntervalCfg = zoomIntervalCfg

    def createFields(self):
        if self["/have_debug"].value:
            yield String(self, "signature", 16)
            if self['signature'].value != "+++IndexStart+++":
                raise ValueError

        indexEntries = []
        numTiles = None
        i = 0
        while True:
            entry = TileIndexEntry(self, "tile_index_entry[]")
            indexEntries.append(entry)
            yield entry

            i += 1
            if numTiles is None:
                # calculate number of tiles (TODO: better calc this from map
                # bounding box)
                firstOffset = self["tile_index_entry[0]"]["offset"].value
                if self["/have_debug"].value:
                    firstOffset -= 16
                numTiles = firstOffset / 5
            if i >= numTiles:
                break

        for i, indexEntry in enumerate(indexEntries):
            offset = indexEntry["offset"].value
            self.seekByte(offset, relative=True)
            if i != len(indexEntries) - 1:
                next_offset = indexEntries[i + 1]["offset"].value
                size = (next_offset - offset) * 8
            else:
                size = self.size - offset * 8
            if size == 0:
                # hachoir doesn't support empty field.
                continue
            yield TileData(self, "tile_data[%d]" % i, zoomIntervalCfg=self.zoomIntervalCfg, size=size)


class MapsforgeMapFile(Parser, RootSeekableFieldSet):
    PARSER_TAGS = {
        "id": "mapsforge_map",
        "category": "misc",
        "file_ext": ("map",),
        "min_size": 62 * 8,
        "description": "Mapsforge map file",
    }

    endian = BIG_ENDIAN

    def validate(self):
        return self["file_magic"].value == "mapsforge binary OSM" and self["file_version"].value == 3

    def createFields(self):
        yield String(self, "file_magic", 20)
        yield UInt32(self, "header_size")
        yield UInt32(self, "file_version")
        yield UInt64(self, "file_size")
        yield UInt64(self, "creation_date")
        yield Int32(self, "min_lat")
        yield Int32(self, "min_lon")
        yield Int32(self, "max_lat")
        yield Int32(self, "max_lon")
        yield UInt16(self, "tile_size")
        yield VbeString(self, "projection")

        # flags
        yield Bit(self, "have_debug")
        yield Bit(self, "have_map_start")
        yield Bit(self, "have_start_zoom")
        yield Bit(self, "have_language_preference")
        yield Bit(self, "have_comment")
        yield Bit(self, "have_created_by")
        yield Bits(self, "reserved[]", 2)

        if self["have_map_start"].value:
            yield UInt32(self, "start_lat")
            yield UInt32(self, "start_lon")
        if self["have_start_zoom"].value:
            yield UInt8(self, "start_zoom")
        if self["have_language_preference"].value:
            yield VbeString(self, "language_preference")
        if self["have_comment"].value:
            yield VbeString(self, "comment")
        if self["have_created_by"].value:
            yield VbeString(self, "created_by")

        yield TagStringList(self, "poi_tags")
        yield TagStringList(self, "way_tags")

        yield UInt8(self, "num_zoom_intervals")
        for i in range(self["num_zoom_intervals"].value):
            yield ZoomIntervalCfg(self, "zoom_interval_cfg[]")

        for i in range(self["num_zoom_intervals"].value):
            zoomIntervalCfg = self["zoom_interval_cfg[%d]" % i]
            self.seekByte(zoomIntervalCfg[
                          "subfile_start"].value, relative=False)
            yield ZoomSubFile(self, "subfile[]", size=zoomIntervalCfg["subfile_size"].value * 8, zoomIntervalCfg=zoomIntervalCfg)



from hachoir.parser.program.elf import ElfFile  # noqa
from hachoir.parser.program.exe import ExeFile  # noqa
from hachoir.parser.program.macho import MachoFile, MachoFatFile  # noqa
from hachoir.parser.program.python import PythonCompiledFile  # noqa
from hachoir.parser.program.java import JavaCompiledClassFile  # noqa
from hachoir.parser.program.prc import PRCFile  # noqa
from hachoir.parser.program.nds import NdsFile  # noqa
from hachoir.parser.program.java_serialized import JavaSerializedFile  # noqa



"""
Tcpdump parser

Source:
 * libpcap source code (file savefile.c)
 * RFC 791 (IPv4)
 * RFC 792 (ICMP)
 * RFC 793 (TCP)
 * RFC 1122 (Requirements for Internet Hosts)

Author: Victor Stinner
Creation: 23 march 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, ParserError,
                           Enum, Bytes, NullBytes, RawBytes,
                           UInt8, UInt16, UInt32, Int32, TimestampUnix32,
                           Bit, Bits, NullBits)
from hachoir.core.endian import NETWORK_ENDIAN, LITTLE_ENDIAN
from hachoir.core.tools import humanDuration
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.tools import createDict
from hachoir.parser.network.common import MAC48_Address, IPv4_Address, IPv6_Address


def diff(field):
    return humanDuration(field.value * 1000)


class Layer(FieldSet):
    endian = NETWORK_ENDIAN

    def parseNext(self, parent):
        return None


class ARP(Layer):
    opcode_name = {
        1: "request",
        2: "reply"
    }
    endian = NETWORK_ENDIAN

    def createFields(self):
        yield UInt16(self, "hw_type")
        yield UInt16(self, "proto_type")
        yield UInt8(self, "hw_size")
        yield UInt8(self, "proto_size")
        yield Enum(UInt16(self, "opcode"), ARP.opcode_name)
        yield MAC48_Address(self, "src_mac")
        yield IPv4_Address(self, "src_ip")
        yield MAC48_Address(self, "dst_mac")
        yield IPv4_Address(self, "dst_ip")

    def createDescription(self):
        desc = "ARP: %s" % self["opcode"].display
        opcode = self["opcode"].value
        src_ip = self["src_ip"].display
        dst_ip = self["dst_ip"].display
        if opcode == 1:
            desc += ", %s ask %s" % (dst_ip, src_ip)
        elif opcode == 2:
            desc += " from %s" % src_ip
        return desc


class TCP_Option(FieldSet):
    NOP = 1
    MAX_SEGMENT = 2
    WINDOW_SCALE = 3
    SACK = 4
    TIMESTAMP = 8

    code_name = {
        NOP: "NOP",
        MAX_SEGMENT: "Max segment size",
        WINDOW_SCALE: "Window scale",
        SACK: "SACK permitted",
        TIMESTAMP: "Timestamp"
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        if self["code"].value != self.NOP:
            self._size = self["length"].value * 8
        else:
            self._size = 8

    def createFields(self):
        yield Enum(UInt8(self, "code", "Code"), self.code_name)
        code = self["code"].value
        if code == self.NOP:
            return
        yield UInt8(self, "length", "Option size in bytes")
        if code == self.MAX_SEGMENT:
            yield UInt16(self, "max_seg", "Maximum segment size")
        elif code == self.WINDOW_SCALE:
            yield UInt8(self, "win_scale", "Window scale")
        elif code == self.TIMESTAMP:
            yield UInt32(self, "ts_val", "Timestamp value")
            yield UInt32(self, "ts_ecr", "Timestamp echo reply")
        else:
            size = (self.size - self.current_size) // 8
            if size:
                yield RawBytes(self, "data", size)

    def createDescription(self):
        return "TCP option: %s" % self["code"].display


class TCP(Layer):
    port_name = {
        13: "daytime",
        20: "ftp data",
        21: "ftp",
        23: "telnet",
        25: "smtp",
        53: "dns",
        63: "dhcp/bootp",
        80: "HTTP",
        110: "pop3",
        119: "nntp",
        123: "ntp",
        139: "netbios session service",
        1863: "MSNMS",
        6667: "IRC"
    }

    def createFields(self):
        yield Enum(UInt16(self, "src"), self.port_name)
        yield Enum(UInt16(self, "dst"), self.port_name)
        yield UInt32(self, "seq_num")
        yield UInt32(self, "ack_num")

        yield Bits(self, "hdrlen", 6, "Header lenght")
        yield NullBits(self, "reserved", 2, "Reserved")

        yield Bit(self, "cgst", "Congestion Window Reduced")
        yield Bit(self, "ecn-echo", "ECN-echo")
        yield Bit(self, "urg", "Urgent")
        yield Bit(self, "ack", "Acknowledge")
        yield Bit(self, "psh", "Push mmode")
        yield Bit(self, "rst", "Reset connection")
        yield Bit(self, "syn", "Synchronize")
        yield Bit(self, "fin", "Stop the connection")

        yield UInt16(self, "winsize", "Windows size")
        yield textHandler(UInt16(self, "checksum"), hexadecimal)
        yield UInt16(self, "urgent")

        size = self["hdrlen"].value * 8 - self.current_size
        while 0 < size:
            option = TCP_Option(self, "option[]")
            yield option
            size -= option.size

    def parseNext(self, parent):
        return None

    def createDescription(self):
        src = self["src"].value
        dst = self["dst"].value
        if src < 32768:
            src = self["src"].display
        else:
            src = None
        if dst < 32768:
            dst = self["dst"].display
        else:
            dst = None
        desc = "TCP"
        if src is not None and dst is not None:
            desc += " (%s->%s)" % (src, dst)
        elif src is not None:
            desc += " (%s->)" % (src)
        elif dst is not None:
            desc += " (->%s)" % (dst)

        # Get flags
        flags = []
        if self["syn"].value:
            flags.append("SYN")
        if self["ack"].value:
            flags.append("ACK")
        if self["fin"].value:
            flags.append("FIN")
        if self["rst"].value:
            flags.append("RST")
        if flags:
            desc += " [%s]" % (",".join(flags))
        return desc


class UDP(Layer):
    port_name = {
        12: "daytime",
        22: "ssh",
        53: "DNS",
        67: "dhcp/bootp",
        80: "http",
        110: "pop3",
        123: "ntp",
        137: "netbios name service",
        138: "netbios datagram service"
    }

    def createFields(self):
        yield Enum(UInt16(self, "src"), UDP.port_name)
        yield Enum(UInt16(self, "dst"), UDP.port_name)
        yield UInt16(self, "length")
        yield textHandler(UInt16(self, "checksum"), hexadecimal)

    def createDescription(self):
        return "UDP (%s->%s)" % (self["src"].display, self["dst"].display)


class ICMP(Layer):
    REJECT = 3
    PONG = 0
    PING = 8
    type_desc = {
        PONG: "Pong",
        REJECT: "Reject",
        PING: "Ping"
    }
    reject_reason = {
        0: "net unreachable",
        1: "host unreachable",
        2: "protocol unreachable",
        3: "port unreachable",
        4: "fragmentation needed and DF set",
        5: "source route failed",
        6: "Destination network unknown error",
        7: "Destination host unknown error",
        8: "Source host isolated error",
        9: "Destination network administratively prohibited",
        10: "Destination host administratively prohibited",
        11: "Unreachable network for Type Of Service",
        12: "Unreachable host for Type Of Service.",
        13: "Communication administratively prohibited",
        14: "Host precedence violation",
        15: "Precedence cutoff in effect"
    }

    def createFields(self):
        # Type
        yield Enum(UInt8(self, "type"), self.type_desc)
        type = self["type"].value

        # Code
        field = UInt8(self, "code")
        if type == 3:
            field = Enum(field, self.reject_reason)
        yield field

        # Options
        yield textHandler(UInt16(self, "checksum"), hexadecimal)
        if type in (self.PING, self.PONG):  # and self["code"].value == 0:
            yield UInt16(self, "id")
            yield UInt16(self, "seq_num")
            # follow: ping data
        elif type == self.REJECT:
            yield NullBytes(self, "empty", 2)
            yield UInt16(self, "hop_mtu", "Next-Hop MTU")

    def createDescription(self):
        type = self["type"].value
        if type in (self.PING, self.PONG):
            return "%s (num=%s)" % (self["type"].display, self["seq_num"].value)
        else:
            return "ICMP (%s)" % self["type"].display

    def parseNext(self, parent):
        if self["type"].value == self.REJECT:
            return IPv4(parent, "rejected_ipv4")
        else:
            return None


class ICMPv6(Layer):
    ECHO_REQUEST = 128
    ECHO_REPLY = 129
    TYPE_DESC = {
        128: "Echo request",
        129: "Echo reply",
    }

    def createFields(self):
        yield Enum(UInt8(self, "type"), self.TYPE_DESC)
        yield UInt8(self, "code")
        yield textHandler(UInt16(self, "checksum"), hexadecimal)

        if self['type'].value in (self.ECHO_REQUEST, self.ECHO_REPLY):
            yield UInt16(self, "id")
            yield UInt16(self, "sequence")

    def createDescription(self):
        if self['type'].value in (self.ECHO_REQUEST, self.ECHO_REPLY):
            return "%s (num=%s)" % (self["type"].display, self["sequence"].value)
        else:
            return "ICMPv6 (%s)" % self["type"].display


class IP(Layer):
    PROTOCOL_INFO = {
        1: ("icmp", ICMP, "ICMP"),
        6: ("tcp", TCP, "TCP"),
        17: ("udp", UDP, "UDP"),
        58: ("icmpv6", ICMPv6, "ICMPv6"),
        60: ("ipv6_opts", None, "IPv6 destination option"),
    }
    PROTOCOL_NAME = createDict(PROTOCOL_INFO, 2)

    def parseNext(self, parent):
        proto = self["protocol"].value
        if proto not in self.PROTOCOL_INFO:
            return None
        name, parser, desc = self.PROTOCOL_INFO[proto]
        if not parser:
            return None
        return parser(parent, name)


class IPv4(IP):
    precedence_name = {
        7: "Network Control",
        6: "Internetwork Control",
        5: "CRITIC/ECP",
        4: "Flash Override",
        3: "Flash",
        2: "Immediate",
        1: "Priority",
        0: "Routine",
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = self["hdr_size"].value * 32

    def createFields(self):
        yield Bits(self, "version", 4, "Version")
        yield Bits(self, "hdr_size", 4, "Header size divided by 5")

        # Type of service
        yield Enum(Bits(self, "precedence", 3, "Precedence"), self.precedence_name)
        yield Bit(self, "low_delay", "If set, low delay, else normal delay")
        yield Bit(self, "high_throu", "If set, high throughput, else normal throughput")
        yield Bit(self, "high_rel", "If set, high relibility, else normal")
        yield NullBits(self, "reserved[]", 2, "(reserved for future use)")

        yield UInt16(self, "length")
        yield UInt16(self, "id")

        yield NullBits(self, "reserved[]", 1)
        yield Bit(self, "df", "Don't fragment")
        yield Bit(self, "more_frag", "There are more fragments? if not set, it's the last one")
        yield Bits(self, "frag_ofst_lo", 5)
        yield UInt8(self, "frag_ofst_hi")
        yield UInt8(self, "ttl", "Type to live")
        yield Enum(UInt8(self, "protocol"), self.PROTOCOL_NAME)
        yield textHandler(UInt16(self, "checksum"), hexadecimal)
        yield IPv4_Address(self, "src")
        yield IPv4_Address(self, "dst")

        size = (self.size - self.current_size) // 8
        if size:
            yield RawBytes(self, "options", size)

    def createDescription(self):
        return "IPv4 (%s>%s)" % (self["src"].display, self["dst"].display)


class IPv6(IP):
    static_size = 40 * 8
    endian = NETWORK_ENDIAN

    def createFields(self):
        yield Bits(self, "version", 4, "Version (6)")
        yield Bits(self, "traffic", 8, "Traffic class")
        yield Bits(self, "flow", 20, "Flow label")
        yield Bits(self, "length", 16, "Payload length")
        yield Enum(Bits(self, "protocol", 8, "Next header"), self.PROTOCOL_NAME)
        yield Bits(self, "hop_limit", 8, "Hop limit")
        yield IPv6_Address(self, "src")
        yield IPv6_Address(self, "dst")

    def createDescription(self):
        return "IPv6 (%s>%s)" % (self["src"].display, self["dst"].display)


class Layer2(Layer):
    PROTO_INFO = {
        0x0800: ("ipv4", IPv4, "IPv4"),
        0x0806: ("arp", ARP, "ARP"),
        0x86dd: ("ipv6", IPv6, "IPv6"),
    }
    PROTO_DESC = createDict(PROTO_INFO, 2)

    def parseNext(self, parent):
        try:
            name, parser, desc = self.PROTO_INFO[self["protocol"].value]
            return parser(parent, name)
        except KeyError:
            return None


class Unicast(Layer2):
    packet_type_name = {
        0: "Unicast to us"
    }

    def createFields(self):
        yield Enum(UInt16(self, "packet_type"), self.packet_type_name)
        yield UInt16(self, "addr_type", "Link-layer address type")
        yield UInt16(self, "addr_length", "Link-layer address length")
        length = self["addr_length"].value
        length = 8   # FIXME: Should we use addr_length or not?
        if length:
            yield RawBytes(self, "source", length)
        yield Enum(UInt16(self, "protocol"), self.PROTO_DESC)


class Ethernet(Layer2):
    static_size = 14 * 8

    def createFields(self):
        yield MAC48_Address(self, "dst")
        yield MAC48_Address(self, "src")
        yield Enum(UInt16(self, "protocol"), self.PROTO_DESC)

    def createDescription(self):
        return "Ethernet: %s>%s (%s)" % \
            (self["src"].display, self["dst"].display, self["protocol"].display)


class Packet(FieldSet):
    endian = LITTLE_ENDIAN

    def __init__(self, parent, name, parser, first_name):
        FieldSet.__init__(self, parent, name)
        self._size = (16 + self["caplen"].value) * 8
        self._first_parser = parser
        self._first_name = first_name

    def createFields(self):
        yield TimestampUnix32(self, "ts_epoch", "Timestamp (Epoch)")
        yield UInt32(self, "ts_nanosec", "Timestamp (nano second)")
        yield UInt32(self, "caplen", "length of portion present")
        yield UInt32(self, "len", "length this packet (off wire)")

        # Read different layers
        field = self._first_parser(self, self._first_name)
        while field:
            yield field
            field = field.parseNext(self)

        # Read data if any
        size = (self.size - self.current_size) // 8
        if size:
            yield RawBytes(self, "data", size)

    def getTimestamp(self):
        nano_sec = float(self["ts_nanosec"].value) / 100
        from datetime import timedelta
        return self["ts_epoch"].value + timedelta(microseconds=nano_sec)

    def createDescription(self):
        t0 = self["/packet[0]"].getTimestamp()
#        ts = max(self.getTimestamp() - t0, t0)
        ts = self.getTimestamp() - t0
        # text = ["%1.6f: " % ts]
        text = ["%s: " % ts]
        if "icmp" in self:
            text.append(self["icmp"].description)
        elif "tcp" in self:
            text.append(self["tcp"].description)
        elif "udp" in self:
            text.append(self["udp"].description)
        elif "arp" in self:
            text.append(self["arp"].description)
        else:
            text.append("Packet")
        return "".join(text)


class TcpdumpFile(Parser):
    PARSER_TAGS = {
        "id": "tcpdump",
        "category": "misc",
        "min_size": 24 * 8,
        "description": "Tcpdump file (network)",
        "magic": ((b"\xd4\xc3\xb2\xa1", 0),),
    }
    endian = LITTLE_ENDIAN

    LINK_TYPE = {
        1: ("ethernet", Ethernet),
        113: ("unicast", Unicast),
    }
    LINK_TYPE_DESC = createDict(LINK_TYPE, 0)

    def validate(self):
        if self["id"].value != b"\xd4\xc3\xb2\xa1":
            return "Wrong file signature"
        if self["link_type"].value not in self.LINK_TYPE:
            return "Unknown link type"
        return True

    def createFields(self):
        yield Bytes(self, "id", 4, "Tcpdump identifier")
        yield UInt16(self, "maj_ver", "Major version")
        yield UInt16(self, "min_ver", "Minor version")
        yield Int32(self, "this_zone", "GMT to local time zone correction")
        yield Int32(self, "sigfigs", "accuracy of timestamps")
        yield UInt32(self, "snap_len", "max length saved portion of each pkt")
        yield Enum(UInt32(self, "link_type", "data link type"), self.LINK_TYPE_DESC)
        link = self["link_type"].value
        if link not in self.LINK_TYPE:
            raise ParserError("Unknown link type: %s" % link)
        name, parser = self.LINK_TYPE[link]
        while self.current_size < self.size:
            yield Packet(self, "packet[]", parser, name)



import hachoir.core.config as config
from hachoir.field import Parser as GenericParser
from hachoir.core.error import error
from hachoir.core.tools import makeUnicode
from inspect import getmro


class ValidateError(Exception):
    pass


class HachoirParser(object):
    """
    A parser is the root of all other fields. It create first level of fields
    and have special attributes and methods:
    - tags: dictionary with keys:
      - "file_ext": classical file extensions (string or tuple of strings) ;
      - "mime": MIME type(s) (string or tuple of strings) ;
      - "description": String describing the parser.
    - endian: Byte order (L{BIG_ENDIAN} or L{LITTLE_ENDIAN}) of input data ;
    - stream: Data input stream (set in L{__init__()}).

    Default values:
    - size: Field set size will be size of input stream ;
    - mime_type: First MIME type of tags["mime"] (if it does exist,
      None otherwise).
    """

    _autofix = False

    def __init__(self, stream, **args):
        validate = args.pop("validate", False)
        self._mime_type = None
        while validate:
            nbits = self.getParserTags()["min_size"]
            if stream.sizeGe(nbits):
                res = self.validate()
                if res is True:
                    break
                res = makeUnicode(res)
            else:
                res = "stream is smaller than %s.%s bytes" % divmod(nbits, 8)
            raise ValidateError(res or "no reason given")
        self._autofix = True

    # --- Methods that can be overridden -------------------------------------
    def createDescription(self):
        """
        Create an Unicode description
        """
        return self.PARSER_TAGS["description"]

    def createMimeType(self):
        """
        Create MIME type (string), eg. "image/png"

        If it returns None, "application/octet-stream" is used.
        """
        if "mime" in self.PARSER_TAGS:
            return self.PARSER_TAGS["mime"][0]
        return None

    def validate(self):
        """
        Check that the parser is able to parse the stream. Valid results:
        - True: stream looks valid ;
        - False: stream is invalid ;
        - str: string describing the error.
        """
        raise NotImplementedError()

    # --- Getter methods -----------------------------------------------------
    def _getDescription(self):
        if self._description is None:
            try:
                self._description = self.createDescription()
                if isinstance(self._description, str):
                    self._description = makeUnicode(self._description)
            except Exception as err:
                error("Error getting description of %s: %s"
                      % (self.path, str(err)))
                self._description = self.PARSER_TAGS["description"]
        return self._description
    description = property(_getDescription,
                           doc="Description of the parser")

    def _getMimeType(self):
        if not self._mime_type:
            try:
                self._mime_type = self.createMimeType()
            except Exception as err:
                self.error("Error when creating MIME type: %s" % str(err))
            if not self._mime_type \
                    and self.createMimeType != Parser.createMimeType:
                self._mime_type = Parser.createMimeType(self)
            if not self._mime_type:
                self._mime_type = "application/octet-stream"
        return self._mime_type
    mime_type = property(_getMimeType)

    def createContentSize(self):
        return None

    def _getContentSize(self):
        if not hasattr(self, "_content_size"):
            try:
                self._content_size = self.createContentSize()
            except Exception as err:
                error("Unable to compute %s content size: %s" %
                      (self.__class__.__name__, err))
                self._content_size = None
        return self._content_size
    content_size = property(_getContentSize)

    def createFilenameSuffix(self):
        """
        Create filename suffix: "." + first value of self.PARSER_TAGS["file_ext"],
        or None if self.PARSER_TAGS["file_ext"] doesn't exist.
        """
        file_ext = self.getParserTags().get("file_ext")
        if isinstance(file_ext, (tuple, list)):
            file_ext = file_ext[0]
        return file_ext and '.' + file_ext

    def _getFilenameSuffix(self):
        if not hasattr(self, "_filename_suffix"):
            self._filename_extension = self.createFilenameSuffix()
        return self._filename_extension
    filename_suffix = property(_getFilenameSuffix)

    @classmethod
    def getParserTags(cls):
        tags = {}
        for cls in reversed(getmro(cls)):
            if hasattr(cls, "PARSER_TAGS"):
                tags.update(cls.PARSER_TAGS)
        return tags

    @classmethod
    def print_(cls, out, verbose):
        tags = cls.getParserTags()
        print("- %s: %s" % (tags["id"], tags["description"]), file=out)
        if verbose:
            if "mime" in tags:
                print("  MIME type: %s" % (", ".join(tags["mime"])), file=out)
            if "file_ext" in tags:
                file_ext = ", ".join(
                    ".%s" % file_ext for file_ext in tags["file_ext"])
                print("  File extension: %s" % file_ext, file=out)

    autofix = property(lambda self: self._autofix and config.autofix)


class Parser(HachoirParser, GenericParser):

    def __init__(self, stream, **args):
        GenericParser.__init__(self, stream)
        HachoirParser.__init__(self, stream, **args)



"""
Extract metadata from RIFF file format: AVI video and WAV sound.
"""

from hachoir.metadata.metadata import (Metadata, MultipleMetadata,
                                       registerExtractor)
from hachoir.metadata.safe import fault_tolerant, getValue
from hachoir.parser.container.riff import RiffFile
from hachoir.parser.video.fourcc import UNCOMPRESSED_AUDIO
from hachoir.core.tools import humanFilesize, makeUnicode, timedelta2seconds
from hachoir.metadata.audio import computeComprRate as computeAudioComprRate
from datetime import timedelta


class RiffMetadata(MultipleMetadata):
    TAG_TO_KEY = {
        "INAM": "title",
        "IART": "artist",
        "ICMT": "comment",
        "ICOP": "copyright",
        "IENG": "author",    # (engineer)
        "ISFT": "producer",
        "ICRD": "creation_date",
        "IDIT": "creation_date",
    }

    def extract(self, riff):
        type = riff["type"].value
        if type == "WAVE":
            self.extractWAVE(riff)
            size = getValue(riff, "audio_data/size")
            if size:
                computeAudioComprRate(self, size * 8)
        elif type == "AVI ":
            if "headers" in riff:
                self.extractAVI(riff["headers"])
                self.extractInfo(riff["headers"])
        elif type == "ACON":
            self.extractAnim(riff)
        if "info" in riff:
            self.extractInfo(riff["info"])

    def processChunk(self, chunk):
        if "text" not in chunk:
            return
        value = chunk["text"].value
        tag = chunk["tag"].value
        if tag not in self.TAG_TO_KEY:
            self.warning("Skip RIFF metadata %s: %s" % (tag, value))
            return
        key = self.TAG_TO_KEY[tag]
        setattr(self, key, value)

    @fault_tolerant
    def extractWAVE(self, wav):
        format = wav["format"]

        # Number of channel, bits/sample, sample rate
        self.nb_channel = format["nb_channel"].value
        self.bits_per_sample = format["bit_per_sample"].value
        self.sample_rate = format["sample_per_sec"].value

        self.compression = format["codec"].display
        if ("nb_sample/nb_sample" in wav
                and 0 < format["sample_per_sec"].value):
            self.duration = timedelta(seconds=float(
                wav["nb_sample/nb_sample"].value) / format["sample_per_sec"].value)
        if format["codec"].value in UNCOMPRESSED_AUDIO:
            # Codec with fixed bit rate
            self.bit_rate = format[
                "nb_channel"].value * format["bit_per_sample"].value * format["sample_per_sec"].value
            if (not self.has("duration")
                and "audio_data/size" in wav
                    and self.has("bit_rate")):
                duration = float(wav["audio_data/size"].value) * \
                    8 / self.get('bit_rate')
                self.duration = timedelta(seconds=duration)

    def extractInfo(self, fieldset):
        for field in fieldset:
            if not field.is_field_set:
                continue
            if "tag" in field:
                if field["tag"].value == "LIST":
                    self.extractInfo(field)
                else:
                    self.processChunk(field)

    @fault_tolerant
    def extractAVIVideo(self, header, meta):
        meta.compression = "%s (fourcc:\"%s\")" \
            % (header["fourcc"].display, makeUnicode(header["fourcc"].value))
        if header["rate"].value and header["scale"].value:
            fps = float(header["rate"].value) / header["scale"].value
            meta.frame_rate = fps
            if 0 < fps:
                self.duration = meta.duration = timedelta(
                    seconds=float(header["length"].value) / fps)

        if "../stream_fmt/width" in header:
            format = header["../stream_fmt"]
            meta.width = format["width"].value
            meta.height = format["height"].value
            meta.bits_per_pixel = format["depth"].value
        else:
            meta.width = header["right"].value - header["left"].value
            meta.height = header["bottom"].value - header["top"].value

    @fault_tolerant
    def extractAVIAudio(self, format, meta):
        meta.nb_channel = format["channel"].value
        meta.sample_rate = format["sample_rate"].value
        meta.bit_rate = format["bit_rate"].value * 8
        if format["bits_per_sample"].value:
            meta.bits_per_sample = format["bits_per_sample"].value
        if "../stream_hdr" in format:
            header = format["../stream_hdr"]
            if header["rate"].value and header["scale"].value:
                frame_rate = float(
                    header["rate"].value) / header["scale"].value
                meta.duration = timedelta(seconds=float(
                    header["length"].value) / frame_rate)
            if header["fourcc"].value != "":
                meta.compression = "%s (fourcc:\"%s\")" \
                    % (format["codec"].display, header["fourcc"].value)
        if not meta.has("compression"):
            meta.compression = format["codec"].display

        self.computeAudioComprRate(meta)

    @fault_tolerant
    def computeAudioComprRate(self, meta):
        uncompr = meta.get('bit_rate', 0)
        if not uncompr:
            return
        compr = meta.get('nb_channel') * meta.get('sample_rate') * \
            meta.get('bits_per_sample', default=16)
        if not compr:
            return
        meta.compr_rate = float(compr) / uncompr

    @fault_tolerant
    def useAviHeader(self, header):
        microsec = header["microsec_per_frame"].value
        if microsec:
            self.frame_rate = 1000000.0 / microsec
            total_frame = getValue(header, "total_frame")
            if total_frame and not self.has("duration"):
                self.duration = timedelta(microseconds=total_frame * microsec)
        self.width = header["width"].value
        self.height = header["height"].value

    def extractAVI(self, headers):
        audio_index = 1
        for stream in headers.array("stream"):
            if "stream_hdr/stream_type" not in stream:
                continue
            stream_type = stream["stream_hdr/stream_type"].value
            if stream_type == "vids":
                if "stream_hdr" in stream:
                    meta = Metadata(self)
                    self.extractAVIVideo(stream["stream_hdr"], meta)
                    self.addGroup("video", meta, "Video stream")
            elif stream_type == "auds":
                if "stream_fmt" in stream:
                    meta = Metadata(self)
                    self.extractAVIAudio(stream["stream_fmt"], meta)
                    self.addGroup("audio[%u]" % audio_index,
                                  meta, "Audio stream")
                    audio_index += 1
        if "avi_hdr" in headers:
            self.useAviHeader(headers["avi_hdr"])

        # Compute global bit rate
        if self.has("duration") and "/movie/size" in headers:
            self.bit_rate = float(
                headers["/movie/size"].value) * 8 / timedelta2seconds(self.get('duration'))

        # Video has index?
        if "/index" in headers:
            self.comment = "Has audio/video index (%s)" \
                % humanFilesize(headers["/index"].size // 8)

    @fault_tolerant
    def extractAnim(self, riff):
        if "anim_rate/rate[0]" in riff:
            count = 0
            total = 0
            for rate in riff.array("anim_rate/rate"):
                count += 1
                if 100 < count:
                    break
                total += rate.value / 60.0
            if count and total:
                self.frame_rate = count / total
        if not self.has("frame_rate") and "anim_hdr/jiffie_rate" in riff:
            self.frame_rate = 60.0 / riff["anim_hdr/jiffie_rate"].value


registerExtractor(RiffFile, RiffMetadata)



from hachoir.field import FieldSet, ParserError


class StaticFieldSet(FieldSet):
    """
    Static field set: format class attribute is a tuple of all fields
    in syntax like:
       format = (
          (TYPE1, ARG1, ARG2, ...),
          (TYPE2, ARG1, ARG2, ..., {KEY1=VALUE1, ...}),
          ...
       )

    Types with dynamic size are forbidden, eg. CString, PascalString8, etc.
    """
    format = None  # You have to redefine this class variable
    _class = None

    def __new__(cls, *args, **kw):
        assert cls.format is not None, "Class attribute 'format' is not set"
        if cls._class is not cls.__name__:
            cls._class = cls.__name__
            cls.static_size = cls._computeStaticSize()
        return object.__new__(cls)

    @staticmethod
    def _computeItemSize(item):
        item_class = item[0]
        if item_class.static_size is None:
            raise ParserError("Unable to get static size of field type: %s"
                              % item_class.__name__)
        if callable(item_class.static_size):
            if isinstance(item[-1], dict):
                return item_class.static_size(*item[1:-1], **item[-1])
            else:
                return item_class.static_size(*item[1:])
        else:
            assert isinstance(item_class.static_size, int)
            return item_class.static_size

    def createFields(self):
        for item in self.format:
            if isinstance(item[-1], dict):
                yield item[0](self, *item[1:-1], **item[-1])
            else:
                yield item[0](self, *item[1:])

    @classmethod
    def _computeStaticSize(cls, *args):
        return sum(cls._computeItemSize(item) for item in cls.format)

    # Initial value of static_size, it changes when first instance
    # is created (see __new__)
    static_size = _computeStaticSize



"""
SWF (Macromedia/Adobe Flash) file parser.

Documentation:

 - Alexis' SWF Reference:
   http://www.m2osw.com/swf_alexref.html
 - Tamarin ABC format:
   http://www.m2osw.com/abc_format.html

Authors: Sebastien Ponce, Robert Xiao
Creation date: 26 April 2008
"""

from hachoir.field import (FieldSet, ParserError,
                           Bit, Bits, UInt8, UInt32, Int16, UInt16, Float32, Float64, CString, Enum,
                           RawBytes, String)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.field.float import FloatExponent
from struct import unpack


class FlashPackedInteger(Bits):

    def __init__(self, parent, name, signed=False, nbits=30, description=None):
        Bits.__init__(self, parent, name, 8, description)
        stream = self._parent.stream
        addr = self.absolute_address
        size = 0
        value = 0
        mult = 1
        while True:
            byte = stream.readBits(addr + size, 8, LITTLE_ENDIAN)
            value += mult * (byte & 0x7f)
            size += 8
            mult <<= 7
            if byte < 128:
                break
        self._size = size
        if signed and (1 << (nbits - 1)) <= value:
            value -= (1 << nbits)
        self.createValue = lambda: value


class FlashU30(FlashPackedInteger):

    def __init__(self, parent, name, description=None):
        FlashPackedInteger.__init__(
            self, parent, name, signed=False, nbits=30, description=description)


class FlashS32(FlashPackedInteger):

    def __init__(self, parent, name, description=None):
        FlashPackedInteger.__init__(
            self, parent, name, signed=True, nbits=32, description=description)


class FlashU32(FlashPackedInteger):

    def __init__(self, parent, name, description=None):
        FlashPackedInteger.__init__(
            self, parent, name, signed=False, nbits=32, description=description)


class FlashFloat64(FieldSet):

    def createFields(self):
        yield Bits(self, "mantissa_high", 20)
        yield FloatExponent(self, "exponent", 11)
        yield Bit(self, "negative")
        yield Bits(self, "mantissa_low", 32)

    def createValue(self):
        # Manual computation:
        # mantissa = mantissa_high * 2^32 + mantissa_low
        # float = 2^exponent + (1 + mantissa / 2^52)
        # (and float is negative if negative=True)
        bytes = self.parent.stream.readBytes(
            self.absolute_address, self.size // 8)
        # Mix bytes: xxxxyyyy <=> yyyyxxxx
        bytes = bytes[4:8] + bytes[0:4]
        return unpack('<d', bytes)[0]


TYPE_INFO = {
    0x00: (CString, "Cstring[]"),
    0x01: (Float32, "Float[]"),
    0x02: (None, "Null[]"),
    0x03: (None, "Undefined[]"),
    0x04: (UInt8, "Register[]"),
    0x05: (UInt8, "Boolean[]"),
    0x06: (FlashFloat64, "Double[]"),
    0x07: (UInt32, "Integer[]"),
    0x08: (UInt8, "Dictionary_Lookup_Index[]"),
    0x09: (UInt16, "Large_Dictionary_Lookup_Index[]"),
}


def parseBranch(parent, size):
    yield Int16(parent, "offset")


def parseDeclareFunction(parent, size):
    yield CString(parent, "name")
    argCount = UInt16(parent, "arg_count")
    yield argCount
    for i in range(argCount.value):
        yield CString(parent, "arg[]")
    yield UInt16(parent, "function_length")


def parseDeclareFunctionV7(parent, size):
    yield CString(parent, "name")
    argCount = UInt16(parent, "arg_count")
    yield argCount
    yield UInt8(parent, "reg_count")
    yield Bits(parent, "reserved", 7)
    yield Bit(parent, "preload_global")
    yield Bit(parent, "preload_parent")
    yield Bit(parent, "preload_root")
    yield Bit(parent, "suppress_super")
    yield Bit(parent, "preload_super")
    yield Bit(parent, "suppress_arguments")
    yield Bit(parent, "preload_arguments")
    yield Bit(parent, "suppress_this")
    yield Bit(parent, "preload_this")
    for i in range(argCount.value):
        yield UInt8(parent, "register[]")
        yield CString(parent, "arg[]")
    yield UInt16(parent, "function_length")


def parseTry(parent, size):
    yield Bits(parent, "reserved", 5)
    catchInReg = Bit(parent, "catch_in_register")
    yield catchInReg
    yield Bit(parent, "finally")
    yield Bit(parent, "catch")
    yield UInt8(parent, "try_size")
    yield UInt8(parent, "catch_size")
    yield UInt8(parent, "finally_size")
    if catchInReg.value:
        yield CString(parent, "name")
    else:
        yield UInt8(parent, "register")


def parsePushData(parent, size):
    while not parent.eof:
        codeobj = UInt8(parent, "data_type[]")
        yield codeobj
        code = codeobj.value
        if code not in TYPE_INFO:
            raise ParserError("Unknown type in Push_Data : " + hex(code))
        parser, name = TYPE_INFO[code]
        if parser:
            yield parser(parent, name)


def parseSetTarget(parent, size):
    yield CString(parent, "target")


def parseWith(parent, size):
    yield UInt16(parent, "size")


def parseGetURL(parent, size):
    yield CString(parent, "url")
    yield CString(parent, "target")


def parseGetURL2(parent, size):
    yield UInt8(parent, "method")


def parseGotoExpression(parent, size):
    yield UInt8(parent, "play")


def parseGotoFrame(parent, size):
    yield UInt16(parent, "frame_no")


def parseGotoLabel(parent, size):
    yield CString(parent, "label")


def parseWaitForFrame(parent, size):
    yield UInt16(parent, "frame")
    yield UInt8(parent, "skip")


def parseWaitForFrameDyn(parent, size):
    yield UInt8(parent, "skip")


def parseDeclareDictionary(parent, size):
    count = UInt16(parent, "count")
    yield count
    for i in range(count.value):
        yield CString(parent, "dictionnary[]")


def parseStoreRegister(parent, size):
    yield UInt8(parent, "register")


def parseStrictMode(parent, size):
    yield UInt8(parent, "strict")


class Instruction(FieldSet):
    ACTION_INFO = {
        0x00: ("end[]", "End", None),
        0x99: ("Branch_Always[]", "Branch Always", parseBranch),
        0x9D: ("Branch_If_True[]", "Branch If True", parseBranch),
        0x3D: ("Call_Function[]", "Call Function", None),
        0x52: ("Call_Method[]", "Call Method", None),
        0x9B: ("Declare_Function[]", "Declare Function", parseDeclareFunction),
        0x8E: ("Declare_Function_V7[]", "Declare Function (V7)", parseDeclareFunctionV7),
        0x3E: ("Return[]", "Return", None),
        0x2A: ("Throw[]", "Throw", None),
        0x8F: ("Try[]", "Try", parseTry),
        # Stack Control
        0x4C: ("Duplicate[]", "Duplicate", None),
        0x96: ("Push_Data[]", "Push Data", parsePushData),
        0x4D: ("Swap[]", "Swap", None),
        # Action Script Context
        0x8B: ("Set_Target[]", "Set Target", parseSetTarget),
        0x20: ("Set_Target_dynamic[]", "Set Target (dynamic)", None),
        0x94: ("With[]", "With", parseWith),
        # Movie Control
        0x9E: ("Call_Frame[]", "Call Frame", None),
        0x83: ("Get_URL[]", "Get URL", parseGetURL),
        0x9A: ("Get_URL2[]", "Get URL2", parseGetURL2),
        0x9F: ("Goto_Expression[]", "Goto Expression", parseGotoExpression),
        0x81: ("Goto_Frame[]", "Goto Frame", parseGotoFrame),
        0x8C: ("Goto_Label[]", "Goto Label", parseGotoLabel),
        0x04: ("Next_Frame[]", "Next Frame", None),
        0x06: ("Play[]", "Play", None),
        0x05: ("Previous_Frame[]", "Previous Frame", None),
        0x07: ("Stop[]", "Stop", None),
        0x08: ("Toggle_Quality[]", "Toggle Quality", None),
        0x8A: ("Wait_For_Frame[]", "Wait For Frame", parseWaitForFrame),
        0x8D: ("Wait_For_Frame_dynamic[]", "Wait For Frame (dynamic)", parseWaitForFrameDyn),
        # Sound
        0x09: ("Stop_Sound[]", "Stop Sound", None),
        # Arithmetic
        0x0A: ("Add[]", "Add", None),
        0x47: ("Add_typed[]", "Add (typed)", None),
        0x51: ("Decrement[]", "Decrement", None),
        0x0D: ("Divide[]", "Divide", None),
        0x50: ("Increment[]", "Increment", None),
        0x18: ("Integral_Part[]", "Integral Part", None),
        0x3F: ("Modulo[]", "Modulo", None),
        0x0C: ("Multiply[]", "Multiply", None),
        0x4A: ("Number[]", "Number", None),
        0x0B: ("Subtract[]", "Subtract", None),
        # Comparisons
        0x0E: ("Equal[]", "Equal", None),
        0x49: ("Equal_typed[]", "Equal (typed)", None),
        0x66: ("Strict_Equal[]", "Strict Equal", None),
        0x67: ("Greater_Than_typed[]", "Greater Than (typed)", None),
        0x0F: ("Less_Than[]", "Less Than", None),
        0x48: ("Less_Than_typed[]", "Less Than (typed)", None),
        0x13: ("String_Equal[]", "String Equal", None),
        0x68: ("String_Greater_Than[]", "String Greater Than", None),
        0x29: ("String_Less_Than[]", "String Less Than", None),
        # Logical and Bit Wise
        0x60: ("And[]", "And", None),
        0x10: ("Logical_And[]", "Logical And", None),
        0x12: ("Logical_Not[]", "Logical Not", None),
        0x11: ("Logical_Or[]", "Logical Or", None),
        0x61: ("Or[]", "Or", None),
        0x63: ("Shift_Left[]", "Shift Left", None),
        0x64: ("Shift_Right[]", "Shift Right", None),
        0x65: ("Shift_Right_Unsigned[]", "Shift Right Unsigned", None),
        0x62: ("Xor[]", "Xor", None),
        # Strings & Characters (See the String Object also)
        0x33: ("Chr[]", "Chr", None),
        0x37: ("Chr_multi-bytes[]", "Chr (multi-bytes)", None),
        0x21: ("Concatenate_Strings[]", "Concatenate Strings", None),
        0x32: ("Ord[]", "Ord", None),
        0x36: ("Ord_multi-bytes[]", "Ord (multi-bytes)", None),
        0x4B: ("String[]", "String", None),
        0x14: ("String_Length[]", "String Length", None),
        0x31: ("String_Length_multi-bytes[]", "String Length (multi-bytes)", None),
        0x15: ("SubString[]", "SubString", None),
        0x35: ("SubString_multi-bytes[]", "SubString (multi-bytes)", None),
        # Properties
        0x22: ("Get_Property[]", "Get Property", None),
        0x23: ("Set_Property[]", "Set Property", None),
        # Objects
        0x2B: ("Cast_Object[]", "Cast Object", None),
        0x42: ("Declare_Array[]", "Declare Array", None),
        0x88: ("Declare_Dictionary[]", "Declare Dictionary", parseDeclareDictionary),
        0x43: ("Declare_Object[]", "Declare Object", None),
        0x3A: ("Delete[]", "Delete", None),
        0x3B: ("Delete_All[]", "Delete All", None),
        0x24: ("Duplicate_Sprite[]", "Duplicate Sprite", None),
        0x46: ("Enumerate[]", "Enumerate", None),
        0x55: ("Enumerate_Object[]", "Enumerate Object", None),
        0x69: ("Extends[]", "Extends", None),
        0x4E: ("Get_Member[]", "Get Member", None),
        0x45: ("Get_Target[]", "Get Target", None),
        0x2C: ("Implements[]", "Implements", None),
        0x54: ("Instance_Of[]", "Instance Of", None),
        0x40: ("New[]", "New", None),
        0x53: ("New_Method[]", "New Method", None),
        0x25: ("Remove_Sprite[]", "Remove Sprite", None),
        0x4F: ("Set_Member[]", "Set Member", None),
        0x44: ("Type_Of[]", "Type Of", None),
        # Variables
        0x41: ("Declare_Local_Variable[]", "Declare Local Variable", None),
        0x1C: ("Get_Variable[]", "Get Variable", None),
        0x3C: ("Set_Local_Variable[]", "Set Local Variable", None),
        0x1D: ("Set_Variable[]", "Set Variable", None),
        # Miscellaneous
        0x2D: ("FSCommand2[]", "FSCommand2", None),
        0x34: ("Get_Timer[]", "Get Timer", None),
        0x30: ("Random[]", "Random", None),
        0x27: ("Start_Drag[]", "Start Drag", None),
        0x28: ("Stop_Drag[]", "Stop Drag", None),
        0x87: ("Store_Register[]", "Store Register", parseStoreRegister),
        0x89: ("Strict_Mode[]", "Strict Mode", parseStrictMode),
        0x26: ("Trace[]", "Trace", None),
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        code = self["action_id"].value
        if code & 128:
            self._size = (3 + self["action_length"].value) * 8
        else:
            self._size = 8
        if code in self.ACTION_INFO:
            self._name, self._description, self.parser = self.ACTION_INFO[code]
        else:
            self.parser = None

    def createFields(self):
        yield Bits(self, "action_id", 8)
        if not (self["action_id"].value & 128):
            return
        yield UInt16(self, "action_length")
        size = self["action_length"].value
        if not size:
            return
        if self.parser:
            yield from self.parser(self, size)
        else:
            yield RawBytes(self, "action_data", size)

    def createDescription(self):
        return self._description

    def __str__(self):
        r = str(self._description)
        for f in self:
            if f.name not in ("action_id", "action_length", "count") and not f.name.startswith("data_type"):
                r = r + "\n   " + \
                    str((self.address + f.address) // 8) + \
                    " " + str(f.name) + "=" + str(f.value)
        return r


class ActionScript(FieldSet):

    def createFields(self):
        while not self.eof:
            yield Instruction(self, "instr[]")

    def __str__(self):
        r = ""
        for f in self:
            r = r + str(f.address // 8) + " " + str(f) + "\n"
        return r


def parseActionScript(parent, size):
    yield ActionScript(parent, "action", size=size * 8)


def FindABC(field):
    while not getattr(field, "isABC", False):
        field = field.parent
        if field is None:
            return None
    return field


def GetConstant(field, pool, index):
    if index == 0:
        return None
    return FindABC(field)["constant_%s_pool/constant[%i]" % (pool, index)]


def GetMultiname(field, index):
    fld = GetConstant(field, "multiname", index)
    if fld is None:
        return "*"
    if "name_index" not in fld:
        return "?"
    fld2 = GetConstant(fld, "string", fld["name_index"].value)
    if fld2 is None:
        return "*"
    return fld2.value


class ABCStringIndex(FlashU30):

    def createDisplay(self):
        fld = GetConstant(self, "string", self.value)
        if fld is None:
            return "*"
        return fld.value


class ABCNSIndex(FlashU30):

    def createDisplay(self):
        fld = GetConstant(self, "namespace", self.value)
        if fld is None:
            return "*"
        return fld.display


class ABCMethodIndex(FlashU30):

    def createDisplay(self):
        fld = FindABC(self)["method_array/method[%i]" % self.value]
        if fld is None:
            return "*"
        return fld.description


class ABCMultinameIndex(FlashU30):

    def createDisplay(self):
        return GetMultiname(self, self.value)


class ABCConstantPool(FieldSet):

    def __init__(self, parent, name, klass):
        FieldSet.__init__(self, parent, 'constant_%s_pool' % name)
        self.klass = klass

    def createFields(self):
        ctr = FlashU30(self, "count")
        yield ctr
        for i in range(ctr.value - 1):
            yield self.klass(self, "constant[%i]" % (i + 1))


class ABCObjectArray(FieldSet):

    def __init__(self, parent, name, klass):
        self.arrname = name
        FieldSet.__init__(self, parent, name + '_array')
        self.klass = klass

    def createFields(self):
        ctr = FlashU30(self, "count")
        yield ctr
        for i in range(ctr.value):
            yield self.klass(self, self.arrname + "[]")


class ABCClassArray(FieldSet):

    def __init__(self, parent, name):
        FieldSet.__init__(self, parent, name + '_array')

    def createFields(self):
        ctr = FlashU30(self, "count")
        yield ctr
        for i in range(ctr.value):
            yield ABCInstanceInfo(self, "instance[]")
        for i in range(ctr.value):
            yield ABCClassInfo(self, "class[]")


class ABCConstantString(FieldSet):

    def createFields(self):
        yield FlashU30(self, "length")
        size = self["length"].value
        if size:
            yield String(self, "data", size, charset="UTF-8")

    def createDisplay(self):
        if "data" in self:
            return self["data"].display
        else:
            return "<empty>"

    def createValue(self):
        if "data" in self:
            return self["data"].value
        else:
            return ""


class ABCConstantNamespace(FieldSet):
    NAMESPACE_KIND = {8: "Namespace",
                      5: "PrivateNamespace",
                      22: "PackageNamespace",
                      23: "PacakgeInternalNamespace",
                      24: "ProtectedNamespace",
                      25: "ExplicitNamespace",
                      26: "MultinameL"}

    def createFields(self):
        yield Enum(UInt8(self, "kind"), self.NAMESPACE_KIND)
        yield ABCStringIndex(self, "name_index")

    def createDisplay(self):
        return "%s %s" % (self["kind"].display, self["name_index"].display)

    def createValue(self):
        return self["name_index"].value


class ABCConstantNamespaceSet(FieldSet):

    def createFields(self):
        ctr = FlashU30(self, "namespace_count")
        yield ctr
        for i in range(ctr.value):
            yield ABCNSIndex(self, "namespace_index[]")

    def createDescription(self):
        ret = [fld.display for fld in self.array("namespace_index")]
        return ', '.join(ret)


class ABCConstantMultiname(FieldSet):
    MULTINAME_KIND = {7: "Qname",
                      13: "QnameA",
                      9: "Multiname",
                      14: "MultinameA",
                      15: "RTQname",
                      16: "RTQnameA",
                      27: "MultinameL",
                      17: "RTQnameL",
                      18: "RTQnameLA"}

    def createFields(self):
        yield Enum(UInt8(self, "kind"), self.MULTINAME_KIND)
        kind = self["kind"].value
        if kind in (7, 13):  # Qname
            yield FlashU30(self, "namespace_index")
            yield ABCStringIndex(self, "name_index")
        elif kind in (9, 14):  # Multiname
            yield ABCStringIndex(self, "name_index")
            yield FlashU30(self, "namespace_set_index")
        elif kind in (15, 16):  # RTQname
            yield ABCStringIndex(self, "name_index")
        elif kind == 27:  # MultinameL
            yield FlashU30(self, "namespace_set_index")
        elif kind in (17, 18):  # RTQnameL
            pass

    def createDisplay(self):
        kind = self["kind"].display
        if "name_index" in self:
            return kind + " " + self["name_index"].display
        return kind

    def createValue(self):
        return self["kind"].value


class ABCTrait(FieldSet):
    TRAIT_KIND = {0: "slot",
                  1: "method",
                  2: "getter",
                  3: "setter",
                  4: "class",
                  5: "function",
                  6: "const", }

    def createFields(self):
        yield ABCMultinameIndex(self, "name_index")
        yield Enum(Bits(self, "kind", 4), self.TRAIT_KIND)
        yield Enum(Bit(self, "is_final"), {True: 'final', False: 'virtual'})
        yield Enum(Bit(self, "is_override"), {True: 'override', False: 'new'})
        yield Bit(self, "has_metadata")
        yield Bits(self, "unused", 1)
        kind = self["kind"].value
        if kind in (0, 6):  # slot, const
            yield FlashU30(self, "slot_id")
            yield ABCMultinameIndex(self, "type_index")
            # TODO reference appropriate constant pool using value_kind
            yield FlashU30(self, "value_index")
            if self['value_index'].value != 0:
                yield UInt8(self, "value_kind")
        elif kind in (1, 2, 3):  # method, getter, setter
            yield FlashU30(self, "disp_id")
            yield ABCMethodIndex(self, "method_info")
        elif kind == 4:  # class
            yield FlashU30(self, "disp_id")
            yield FlashU30(self, "class_info")
        elif kind == 5:  # function
            yield FlashU30(self, "disp_id")
            yield ABCMethodIndex(self, "method_info")
        if self['has_metadata'].value:
            yield ABCObjectArray(self, "metadata", FlashU30)


class ABCValueKind(FieldSet):

    def createFields(self):
        yield FlashU30(self, "value_index")
        yield UInt8(self, "value_kind")


class ABCMethodInfo(FieldSet):

    def createFields(self):
        yield FlashU30(self, "param_count")
        yield ABCMultinameIndex(self, "ret_type")
        for i in range(self["param_count"].value):
            yield ABCMultinameIndex(self, "param_type[]")
        yield ABCStringIndex(self, "name_index")
        yield Bit(self, "need_arguments")
        yield Bit(self, "need_activation")
        yield Bit(self, "need_rest")
        yield Bit(self, "has_optional")
        yield Bit(self, "ignore_rest")
        yield Bit(self, "explicit")
        yield Bit(self, "setsdxns")
        yield Bit(self, "has_paramnames")
        if self["has_optional"].value:
            yield ABCObjectArray(self, "optional", ABCValueKind)
        if self["has_paramnames"].value:
            for i in range(self["param_count"].value):
                yield FlashU30(self, "param_name[]")

    def createDescription(self):
        ret = GetMultiname(self, self["ret_type"].value)
        ret += " " + self["name_index"].display
        ret += "(" + ", ".join(GetMultiname(self, fld.value)
                               for fld in self.array("param_type")) + ")"
        return ret


class ABCMetadataInfo(FieldSet):

    def createFields(self):
        yield ABCStringIndex(self, "name_index")
        yield FlashU30(self, "values_count")
        count = self["values_count"].value
        for i in range(count):
            yield FlashU30(self, "key[]")
        for i in range(count):
            yield FlashU30(self, "value[]")


class ABCInstanceInfo(FieldSet):

    def createFields(self):
        yield ABCMultinameIndex(self, "name_index")
        yield ABCMultinameIndex(self, "super_index")
        yield Bit(self, "is_sealed")
        yield Bit(self, "is_final")
        yield Bit(self, "is_interface")
        yield Bit(self, "is_protected")
        yield Bits(self, "unused", 4)
        if self['is_protected'].value:
            yield ABCNSIndex(self, "protectedNS")
        yield FlashU30(self, "interfaces_count")
        for i in range(self["interfaces_count"].value):
            yield ABCMultinameIndex(self, "interface[]")
        yield ABCMethodIndex(self, "iinit_index")
        yield ABCObjectArray(self, "trait", ABCTrait)


class ABCClassInfo(FieldSet):

    def createFields(self):
        yield ABCMethodIndex(self, "cinit_index")
        yield ABCObjectArray(self, "trait", ABCTrait)


class ABCScriptInfo(FieldSet):

    def createFields(self):
        yield ABCMethodIndex(self, "init_index")
        yield ABCObjectArray(self, "trait", ABCTrait)


class ABCException(FieldSet):

    def createFields(self):
        yield FlashU30(self, "start")
        yield FlashU30(self, "end")
        yield FlashU30(self, "target")
        yield FlashU30(self, "type_index")
        yield FlashU30(self, "name_index")


class ABCMethodBody(FieldSet):

    def createFields(self):
        yield ABCMethodIndex(self, "method_info")
        yield FlashU30(self, "max_stack")
        yield FlashU30(self, "max_regs")
        yield FlashU30(self, "scope_depth")
        yield FlashU30(self, "max_scope")
        yield FlashU30(self, "code_length")
        yield RawBytes(self, "code", self['code_length'].value)
        yield ABCObjectArray(self, "exception", ABCException)
        yield ABCObjectArray(self, "trait", ABCTrait)


def parseABC(parent, size):
    code = parent["code"].value
    if code == parent.TAG_DO_ABC_DEFINE:
        yield UInt32(parent, "action_flags")
        yield CString(parent, "action_name")
    yield UInt16(parent, "minor_version")
    yield UInt16(parent, "major_version")
    parent.isABC = True

    yield ABCConstantPool(parent, "int", FlashS32)
    yield ABCConstantPool(parent, "uint", FlashU32)
    yield ABCConstantPool(parent, "double", Float64)
    yield ABCConstantPool(parent, "string", ABCConstantString)
    yield ABCConstantPool(parent, "namespace", ABCConstantNamespace)
    yield ABCConstantPool(parent, "namespace_set", ABCConstantNamespaceSet)
    yield ABCConstantPool(parent, "multiname", ABCConstantMultiname)

    yield ABCObjectArray(parent, "method", ABCMethodInfo)
    yield ABCObjectArray(parent, "metadata", ABCMetadataInfo)
    yield ABCClassArray(parent, "class")
    yield ABCObjectArray(parent, "script", ABCScriptInfo)
    yield ABCObjectArray(parent, "body", ABCMethodBody)



from hachoir.core import config
from locale import setlocale, LC_ALL


def setup_tests():
    config.use_i18n = False  # Don't use i18n
    config.quiet = True      # Don't display warnings
    setlocale(LC_ALL, "C")



"""
Parser of FastTrackerII Extended Module (XM) version 1.4

Documents:
- Modplug source code (file modplug/soundlib/Load_xm.cpp)
  http://sourceforge.net/projects/modplug
- Dumb source code (files include/dumb.h and src/it/readxm.c
  http://dumb.sf.net/
- Documents of "XM" format on Wotsit
  http://www.wotsit.org

Author: Christophe GISQUET <christophe.gisquet@free.fr>
Creation: 8th February 2007
"""

from hachoir.parser import Parser
from hachoir.field import (StaticFieldSet, FieldSet,
                           Bit, RawBits, Bits,
                           UInt32, UInt16, UInt8, Int8, Enum,
                           RawBytes, String, GenericVector)
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN
from hachoir.core.text_handler import textHandler, filesizeHandler, hexadecimal
from hachoir.parser.audio.modplug import ParseModplugMetadata
from hachoir.parser.common.tracker import NOTE_NAME


def parseSigned(val):
    return "%i" % (val.value - 128)


# From dumb
SEMITONE_BASE = 1.059463094359295309843105314939748495817
PITCH_BASE = 1.000225659305069791926712241547647863626

SAMPLE_LOOP_MODE = ("No loop", "Forward loop", "Ping-pong loop", "Undef")


class SampleType(FieldSet):
    static_size = 8

    def createFields(self):
        yield Bits(self, "unused[]", 4)
        yield Bit(self, "16bits")
        yield Bits(self, "unused[]", 1)
        yield Enum(Bits(self, "loop_mode", 2), SAMPLE_LOOP_MODE)


class SampleHeader(FieldSet):
    static_size = 40 * 8

    def createFields(self):
        yield UInt32(self, "length")
        yield UInt32(self, "loop_start")
        yield UInt32(self, "loop_end")
        yield UInt8(self, "volume")
        yield Int8(self, "fine_tune")
        yield SampleType(self, "type")
        yield UInt8(self, "panning")
        yield Int8(self, "relative_note")
        yield UInt8(self, "reserved")
        yield String(self, "name", 22, charset="ASCII", strip=' \0')

    def createValue(self):
        bytes = 1 + self["type/16bits"].value
        C5_speed = int(16726.0 * pow(SEMITONE_BASE, self["relative_note"].value)
                       * pow(PITCH_BASE, self["fine_tune"].value * 2))
        return "%s, %ubits, %u samples, %uHz" % \
               (self["name"].display, 8 * bytes,
                self["length"].value / bytes, C5_speed)


class StuffType(StaticFieldSet):
    format = (
        (Bits, "unused", 5),
        (Bit, "loop"),
        (Bit, "sustain"),
        (Bit, "on")
    )


class InstrumentSecondHeader(FieldSet):
    static_size = 234 * 8

    def createFields(self):
        yield UInt32(self, "sample_header_size")
        yield GenericVector(self, "notes", 96, UInt8, "sample")
        yield GenericVector(self, "volume_envelope", 24, UInt16, "point")
        yield GenericVector(self, "panning_envelope", 24, UInt16, "point")
        yield UInt8(self, "volume_points", r"Number of volume points")
        yield UInt8(self, "panning_points", r"Number of panning points")
        yield UInt8(self, "volume_sustain_point")
        yield UInt8(self, "volume_loop_start_point")
        yield UInt8(self, "volume_loop_end_point")
        yield UInt8(self, "panning_sustain_point")
        yield UInt8(self, "panning_loop_start_point")
        yield UInt8(self, "panning_loop_end_point")
        yield StuffType(self, "volume_type")
        yield StuffType(self, "panning_type")
        yield UInt8(self, "vibrato_type")
        yield UInt8(self, "vibrato_sweep")
        yield UInt8(self, "vibrato_depth")
        yield UInt8(self, "vibrato_rate")
        yield UInt16(self, "volume_fadeout")
        yield GenericVector(self, "reserved", 11, UInt16, "word")


def createInstrumentContentSize(s, addr):
    start = addr
    samples = s.stream.readBits(addr + 27 * 8, 16, LITTLE_ENDIAN)
    # Seek to end of header (1st + 2nd part)
    addr += 8 * s.stream.readBits(addr, 32, LITTLE_ENDIAN)

    sample_size = 0
    if samples:
        for index in range(samples):
            # Read the sample size from the header
            sample_size += s.stream.readBits(addr, 32, LITTLE_ENDIAN)
            # Seek to next sample header
            addr += SampleHeader.static_size

    return addr - start + 8 * sample_size


class Instrument(FieldSet):

    def __init__(self, parent, name):
        FieldSet.__init__(self, parent, name)
        self._size = createInstrumentContentSize(self, self.absolute_address)
        self.info(self.createDescription())

    # Seems to fix things...
    def fixInstrumentHeader(self):
        size = self["size"].value - self.current_size // 8
        if size:
            yield RawBytes(self, "unknown_data", size)

    def createFields(self):
        yield UInt32(self, "size")
        yield String(self, "name", 22, charset="ASCII", strip=" \0")
        # Doc says type is always 0, but I've found values of 24 and 96 for
        # the _same_ song here, just different download sources for the file
        yield UInt8(self, "type")
        yield UInt16(self, "samples")
        num = self["samples"].value
        self.info(self.createDescription())

        if num:
            yield InstrumentSecondHeader(self, "second_header")

            yield from self.fixInstrumentHeader()

            # This part probably wrong
            sample_size = []
            for index in range(num):
                sample = SampleHeader(self, "sample_header[]")
                yield sample
                sample_size.append(sample["length"].value)

            for size in sample_size:
                if size:
                    yield RawBytes(self, "sample_data[]", size, "Deltas")
        else:
            yield from self.fixInstrumentHeader()

    def createDescription(self):
        return "Instrument '%s': %i samples, header %i bytes" % \
               (self["name"].value, self["samples"].value, self["size"].value)


VOLUME_NAME = (
    "Volume slide down", "Volume slide up", "Fine volume slide down",
    "Fine volume slide up", "Set vibrato speed", "Vibrato",
    "Set panning", "Panning slide left", "Panning slide right",
    "Tone porta", "Unhandled")


def parseVolume(val):
    val = val.value
    if 0x10 <= val <= 0x50:
        return "Volume %i" % val - 16
    else:
        return VOLUME_NAME[val / 16 - 6]


class RealBit(RawBits):
    static_size = 1

    def __init__(self, parent, name, description=None):
        RawBits.__init__(self, parent, name, 1, description=description)

    def createValue(self):
        return self._parent.stream.readBits(self.absolute_address, 1, BIG_ENDIAN)


class NoteInfo(StaticFieldSet):
    format = (
        (RawBits, "unused", 2),
        (RealBit, "has_parameter"),
        (RealBit, "has_type"),
        (RealBit, "has_volume"),
        (RealBit, "has_instrument"),
        (RealBit, "has_note")
    )


EFFECT_NAME = (
    "Arppegio", "Porta up", "Porta down", "Tone porta", "Vibrato",
    "Tone porta+Volume slide", "Vibrato+Volume slide", "Tremolo",
    "Set panning", "Sample offset", "Volume slide", "Position jump",
    "Set volume", "Pattern break", None, "Set tempo/BPM",
    "Set global volume", "Global volume slide", "Unused", "Unused",
    "Unused", "Set envelope position", "Unused", "Unused",
    "Panning slide", "Unused", "Multi retrig note", "Unused",
    "Tremor", "Unused", "Unused", "Unused", None)

EFFECT_E_NAME = (
    "Unknown", "Fine porta up", "Fine porta down",
    "Set gliss control", "Set vibrato control", "Set finetune",
    "Set loop begin/loop", "Set tremolo control", "Retrig note",
    "Fine volume slide up", "Fine volume slide down", "Note cut",
    "Note delay", "Pattern delay")


class Effect(RawBits):

    def __init__(self, parent, name):
        RawBits.__init__(self, parent, name, 8)

    def createValue(self):
        t = self.parent.stream.readBits(
            self.absolute_address, 8, LITTLE_ENDIAN)
        param = self.parent.stream.readBits(
            self.absolute_address + 8, 8, LITTLE_ENDIAN)
        if t == 0x0E:
            return EFFECT_E_NAME[param >> 4] + " %i" % (param & 0x07)
        elif t == 0x21:
            return ("Extra fine porta up", "Extra fine porta down")[param >> 4]
        else:
            return EFFECT_NAME[t]


class Note(FieldSet):

    def __init__(self, parent, name, desc=None):
        FieldSet.__init__(self, parent, name, desc)
        self.flags = self.stream.readBits(
            self.absolute_address, 8, LITTLE_ENDIAN)
        if self.flags & 0x80:
            # TODO: optimize bitcounting with a table:
            # http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetTable
            self._size = 8
            if self.flags & 0x01:
                self._size += 8
            if self.flags & 0x02:
                self._size += 8
            if self.flags & 0x04:
                self._size += 8
            if self.flags & 0x08:
                self._size += 8
            if self.flags & 0x10:
                self._size += 8
        else:
            self._size = 5 * 8

    def createFields(self):
        # This stupid shit gets the LSB, not the MSB...
        self.info("Note info: 0x%02X" %
                  self.stream.readBits(self.absolute_address, 8, LITTLE_ENDIAN))
        yield RealBit(self, "is_extended")
        if self["is_extended"].value:
            info = NoteInfo(self, "info")
            yield info
            if info["has_note"].value:
                yield Enum(UInt8(self, "note"), NOTE_NAME)
            if info["has_instrument"].value:
                yield UInt8(self, "instrument")
            if info["has_volume"].value:
                yield textHandler(UInt8(self, "volume"), parseVolume)
            if info["has_type"].value:
                yield Effect(self, "effect_type")
            if info["has_parameter"].value:
                yield textHandler(UInt8(self, "effect_parameter"), hexadecimal)
        else:
            yield Enum(Bits(self, "note", 7), NOTE_NAME)
            yield UInt8(self, "instrument")
            yield textHandler(UInt8(self, "volume"), parseVolume)
            yield Effect(self, "effect_type")
            yield textHandler(UInt8(self, "effect_parameter"), hexadecimal)

    def createDescription(self):
        if "info" in self:
            info = self["info"]
            desc = []
            if info["has_note"].value:
                desc.append(self["note"].display)
            if info["has_instrument"].value:
                desc.append("instrument %i" % self["instrument"].value)
            if info["has_volume"].value:
                desc.append(self["has_volume"].display)
            if info["has_type"].value:
                desc.append("effect %s" % self["effect_type"].value)
            if info["has_parameter"].value:
                desc.append("parameter %i" % self["effect_parameter"].value)
        else:
            desc = (self["note"].display, "instrument %i" % self["instrument"].value,
                    self["has_volume"].display, "effect %s" % self[
                        "effect_type"].value,
                    "parameter %i" % self["effect_parameter"].value)
        if desc:
            return "Note %s" % ", ".join(desc)
        else:
            return "Note"


class Row(FieldSet):

    def createFields(self):
        for index in range(self["/header/channels"].value):
            yield Note(self, "note[]")


def createPatternContentSize(s, addr):
    return 8 * (s.stream.readBits(addr, 32, LITTLE_ENDIAN) +
                s.stream.readBits(addr + 7 * 8, 16, LITTLE_ENDIAN))


class Pattern(FieldSet):

    def __init__(self, parent, name, desc=None):
        FieldSet.__init__(self, parent, name, desc)
        self._size = createPatternContentSize(self, self.absolute_address)

    def createFields(self):
        yield UInt32(self, "header_size", r"Header length (9)")
        yield UInt8(self, "packing_type", r"Packing type (always 0)")
        yield UInt16(self, "rows", r"Number of rows in pattern (1..256)")
        yield UInt16(self, "data_size", r"Packed patterndata size")
        rows = self["rows"].value
        self.info("Pattern: %i rows" % rows)
        for index in range(rows):
            yield Row(self, "row[]")

    def createDescription(self):
        return "Pattern with %i rows" % self["rows"].value


class Header(FieldSet):
    MAGIC = b"Extended Module: "
    static_size = 336 * 8

    def createFields(self):
        yield String(self, "signature", 17, "XM signature", charset="ASCII")
        yield String(self, "title", 20, "XM title", charset="ASCII", strip=' ')
        yield UInt8(self, "marker", "Marker (0x1A)")
        yield String(self, "tracker_name", 20, "XM tracker name", charset="ASCII", strip=' ')
        yield UInt8(self, "format_minor")
        yield UInt8(self, "format_major")
        yield filesizeHandler(UInt32(self, "header_size", "Header size (276)"))
        yield UInt16(self, "song_length", "Length in patten order table")
        yield UInt16(self, "restart", "Restart position")
        yield UInt16(self, "channels", "Number of channels (2,4,6,8,10,...,32)")
        yield UInt16(self, "patterns", "Number of patterns (max 256)")
        yield UInt16(self, "instruments", "Number of instruments (max 128)")
        yield Bit(self, "amiga_ftable", "Amiga frequency table")
        yield Bit(self, "linear_ftable", "Linear frequency table")
        yield Bits(self, "unused", 14)
        yield UInt16(self, "tempo", "Default tempo")
        yield UInt16(self, "bpm", "Default BPM")
        yield GenericVector(self, "pattern_order", 256, UInt8, "order")

    def createDescription(self):
        return "'%s' by '%s'" % (
            self["title"].value, self["tracker_name"].value)


class XMModule(Parser):
    PARSER_TAGS = {
        "id": "fasttracker2",
        "category": "audio",
        "file_ext": ("xm",),
        "mime": (
            'audio/xm', 'audio/x-xm',
            'audio/module-xm', 'audio/mod', 'audio/x-mod'),
        "magic": ((Header.MAGIC, 0),),
        "min_size": Header.static_size + 29 * 8,  # Header + 1 empty instrument
        "description": "FastTracker2 module"
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        header = self.stream.readBytes(0, 17)
        if header != Header.MAGIC:
            return "Invalid signature %a" % header
        if self["/header/header_size"].value != 276:
            return "Unknown header size (%u)" % self["/header/header_size"].value
        return True

    def createFields(self):
        yield Header(self, "header")
        for index in range(self["/header/patterns"].value):
            yield Pattern(self, "pattern[]")
        for index in range(self["/header/instruments"].value):
            yield Instrument(self, "instrument[]")

        # Metadata added by ModPlug - can be discarded
        yield from ParseModplugMetadata(self)

    def createContentSize(self):
        # Header size
        size = Header.static_size

        # Add patterns size
        for index in range(self["/header/patterns"].value):
            size += createPatternContentSize(self, size)

        # Add instruments size
        for index in range(self["/header/instruments"].value):
            size += createInstrumentContentSize(self, size)

        # Not reporting Modplug metadata
        return size

    def createDescription(self):
        return self["header"].description



import wx


class field_view_fwd_t:
    def __init__(self, imp):
        self.imp = imp

    def on_field_view_ready(self, dispatcher, field_view):
        assert field_view is not None

        field_view.Bind(wx.EVT_COMMAND_RIGHT_CLICK, self.on_item_right_clicked)
        field_view.Bind(wx.EVT_RIGHT_UP, self.on_item_right_clicked)
        field_view.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.on_item_activated)
        field_view.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_item_selected)

    def on_item_activated(self, event):
        self.imp.on_item_activated()

    def on_item_selected(self, event):
        self.imp.on_item_selected()

    def on_item_right_clicked(self, event):
        self.imp.on_item_show_ops()
